-- See the Copyright notice at the end of this file.
--
class PROCESS
	--
	-- This class allows one to spawn an external process and make it execute some file.
	--
	-- The standard streams of the process are available: `input', `output' and `error'.
	--
	-- '''Note:''' This class is in a beta stage. Only the POSIX version is available (Windows implementers
	-- needed!)
	--

creation {ANY}
	execute

feature {ANY}
	id: INTEGER is
		do
			Result := basic_exec_id(data)
		end

	is_connected: BOOLEAN is
		do
			Result := data /= default_pointer
		end

	is_finished: BOOLEAN is
		require
			is_connected
		do
			Result := basic_exec_is_finished(data)
		end

	status: INTEGER is
		require
			is_finished
		do
			Result := basic_exec_status(data)
		end

	wait is
		require
			is_connected
		do
			basic_exec_wait(data)
		ensure
			is_finished
		end

	kill (signal: INTEGER) is
		require
			is_connected
		do
			basic_exec_kill(data, signal)
		end

	input: OUTPUT_STREAM
			-- You can feed data to the process by this channel

	output: INPUT_STREAM
			-- You can read data from the process by this channel

	error: INPUT_STREAM
			-- You can read error data from the process by this channel

feature {}
	execute (program: STRING; arguments: COLLECTION[STRING]; keep_environment: BOOLEAN) is
			-- Execute the given `program' (how the program is discovered is os-dependant), passing to it the
			-- `arguments'. The environment is cleared if `keep_environment' is False.
		require
			program /= Void
		local
			args: FAST_ARRAY[POINTER]; i: INTEGER; ext_args: POINTER; instream: EXEC_OUTPUT_STREAM
			outstream, errstream: EXEC_INPUT_STREAM
		do
			if arguments /= Void then
				from
					create args.with_capacity(arguments.count + 2)
					args.add_last(program.to_external)
					i := arguments.lower
				until
					i > arguments.upper
				loop
					args.add_last(arguments.item(i).to_external)
					i := i + 1
				end
				args.add_last(default_pointer)
				ext_args := args.to_external
			else
				create args.with_capacity(2)
				args.add_last(program.to_external)
				args.add_last(default_pointer)
			end
			create instream.make(Current)
			create outstream.make(Current)
			create errstream.make(Current)
			data := basic_exec_execute(program.to_external, ext_args, keep_environment, instream.pipe, outstream.pipe, errstream.pipe)
			if id > 0 then
				input := instream
				output := outstream
				error := errstream
			end
		end

feature {EXEC_INPUT_STREAM, EXEC_OUTPUT_STREAM} -- plugin lowlevel data
	data: POINTER

feature {} -- plugin features
	basic_exec_execute (program: POINTER; arguments: POINTER; keep_environment: BOOLEAN; in_fd, out_fd, err_fd: POINTER): like data is
		external "plug_in"
		alias "{
			location: "${sys}plugins/exec"
			module_name: "exec"
			feature_name: "basic_exec_execute"
			}"
		end

	basic_exec_id (dat: like data): INTEGER is
		external "plug_in"
		alias "{
			location: "${sys}plugins/exec"
			module_name: "exec"
			feature_name: "basic_exec_id"
			}"
		end

	basic_exec_is_finished (dat: like data): BOOLEAN is
		external "plug_in"
		alias "{
			location: "${sys}plugins/exec"
			module_name: "exec"
			feature_name: "basic_exec_is_finished"
			}"
		end

	basic_exec_status (dat: like data): INTEGER is
		external "plug_in"
		alias "{
			location: "${sys}plugins/exec"
			module_name: "exec"
			feature_name: "basic_exec_status"
			}"
		end

	basic_exec_wait (dat: like data) is
		external "plug_in"
		alias "{
			location: "${sys}plugins/exec"
			module_name: "exec"
			feature_name: "basic_exec_wait"
			}"
		end

	basic_exec_kill (dat: like data; signal: INTEGER) is
		external "plug_in"
		alias "{
			location: "${sys}plugins/exec"
			module_name: "exec"
			feature_name: "basic_exec_kill"
			}"
		end

end -- class PROCESS
--
-- ------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is part of the SmartEiffel standard library.
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2006: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-- documentation files (the "Software"), to deal in the Software without restriction, including without
-- limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
-- the Software, and to permit persons to whom the Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
-- EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
-- AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
-- OR OTHER DEALINGS IN THE SOFTWARE.
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------
