-- See the Copyright notice at the end of this file.
--
class BINARY_FILE_READ
	-- This class allow to read a file on the disk as a binary file
	-- (ie. file containing bytes). If you need to read a file which
	-- contain text, then consider using TEXT_FILE_READ.

inherit
	FILE

creation {ANY}
	make, connect_to

feature {ANY}
	connect_to (new_path: STRING) is
			-- Open binary file for reading. The stream is positioned at the
			-- beginning of the file.
		do
			input_stream := binary_file_read_open(new_path.to_external)
			if input_stream.is_not_null then
				path := new_path
				end_of_input := False
				if capacity = 0 then
					buffer := buffer.calloc(4096)
					capacity := 4096
				end
				buffer_position := 0
				buffer_size := 0
			end
		ensure then
			is_connected implies not end_of_input
		end

	disconnect is
		do
			io_fclose(input_stream)
			path := Void
		end

	read_byte is
			-- Read a byte and assign it to `last_byte'.
		require
			is_connected
			not end_of_input
		do
			if buffer_position >= buffer_size then
				fill_buffer
			end
			last_byte := buffer.item(buffer_position).code
			buffer_position := buffer_position + 1
		end

	last_byte: INTEGER
			-- Last byte read with `read_byte'.

	read_integer_16_native_endian is
			-- Read in the same order as the machine running this code. If a
			-- 16-bits value is available, it's assigned to `last_integer_16'.
			-- The result is machine dependant.
		require
			is_connected
			not end_of_input
		local
			c1, c2: CHARACTER
		do
			if buffer_position >= buffer_size then
				fill_buffer
			end
			c1 := buffer.item(buffer_position)
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			c2 := buffer.item(buffer_position)
			buffer_position := buffer_position + 1
			last_integer_16 := as_16_ne(c1, c2)
		end

	read_integer_16_big_endian is
			-- Read a big endian value is the file. If a 16-bits value
			-- is available, it's assigned to `last_integer_16'.
		require
			is_connected
			not end_of_input
		do
			if buffer_position >= buffer_size then
				fill_buffer
			end
			last_integer_16 := buffer.item(buffer_position).code |<< 8
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			last_integer_16 := last_integer_16 | buffer.item(buffer_position).code
			buffer_position := buffer_position + 1
		end

	read_integer_16_little_endian is
			-- Read a little endian value is the file. If a 16-bits value
			-- is available, it's assigned to `last_integer_16'.
		require
			is_connected
			not end_of_input
		do
			if buffer_position >= buffer_size then
				fill_buffer
			end
			last_integer_16 := buffer.item(buffer_position).code
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			last_integer_16 := last_integer_16 | (buffer.item(buffer_position).code |<< 8)
			buffer_position := buffer_position + 1
		end

	last_integer_16: INTEGER
			-- Last byte read with `read_integer_16_*'.

	read_integer_32_native_endian is
			-- Read in the same order as the machine running this code. If a
			-- 32-bits value is available, it's assigned to `last_integer_32'.
			-- The result is machine dependant.
		require
			is_connected
			not end_of_input
		local
			old_integer_16, i1, i2: INTEGER
		do
			old_integer_16 := last_integer_16
			read_integer_16_native_endian
			i1 := last_integer_16
			read_integer_16_native_endian
			i2 := last_integer_16
			last_integer_32 := as_32_ne(i1.to_integer_16, i2.to_integer_16)
			last_integer_16 := old_integer_16
		end

	read_integer_32_big_endian is
			-- Read a big endian value is the file. If 32-bits value
			-- is available, it's assigned to `last_integer_32'.
		require
			is_connected
			not end_of_input
		do
			if buffer_position >= buffer_size then
				fill_buffer
			end
			last_integer_32 := buffer.item(buffer_position).code |<< 24
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			last_integer_32 := last_integer_32 | (buffer.item(buffer_position).code |<< 16)
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			last_integer_32 := last_integer_32 | (buffer.item(buffer_position).code |<< 8)
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			last_integer_32 := last_integer_32 | buffer.item(buffer_position).code
			buffer_position := buffer_position + 1
		end

	read_integer_32_little_endian is
			-- Read a little endian value is the file. If a 32-bits value
			-- is available, it's assigned to `last_integer_32'.
		require
			is_connected
			not end_of_input
		do
			if buffer_position >= buffer_size then
				fill_buffer
			end
			last_integer_32 := buffer.item(buffer_position).code
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			last_integer_32 := last_integer_32 | (buffer.item(buffer_position).code |<< 8)
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			last_integer_32 := last_integer_32 | (buffer.item(buffer_position).code |<< 16)
			buffer_position := buffer_position + 1
			if buffer_position >= buffer_size and then not end_of_input then
				fill_buffer
			end
			last_integer_32 := last_integer_32 | (buffer.item(buffer_position).code |<< 24)
			buffer_position := buffer_position + 1
		end

	last_integer_32: INTEGER

	end_of_input: BOOLEAN
			-- Has end-of-input been reached ?
			-- True when the last character has been read.

feature {}
	buffer: NATIVE_ARRAY[CHARACTER]

	buffer_position, buffer_size: INTEGER

	capacity: INTEGER

	input_stream: POINTER

	fill_buffer is
		do
			buffer_size := io_fread(buffer, capacity, input_stream)
			buffer_position := 0
			if buffer_size <= 0 then
				end_of_input := True --if buffer_size = -1 => exception ?
				buffer_size := 1
				buffer_position := 1
			end
		end

	make is
			-- The new created object is not connected. (See also `connect_to'.)
		do
		ensure
			not is_connected
		end

	binary_file_read_open (path_pointer: POINTER): POINTER is
		external "plug_in"
		alias "{
			location: "${sys}plugins"
			module_name: "io"
			feature_name: "binary_file_read_open"
			}"
		end

	io_fread (buf: NATIVE_ARRAY[CHARACTER]; size: INTEGER; stream_pointer: POINTER): INTEGER is
			-- return size read or 0 if end of input (-1 on error => exception ?)
		external "plug_in"
		alias "{
			location: "${sys}plugins"
			module_name: "io"
			feature_name: "io_fread"
			}"
		end

	io_fclose (stream_pointer: POINTER) is
		external "plug_in"
		alias "{
			location: "${sys}plugins"
			module_name: "io"
			feature_name: "io_fclose"
			}"
		end

	as_16_ne (c1, c2: CHARACTER): INTEGER_16 is
		external "built_in"
		end

	as_32_ne (i1, i2: INTEGER_16): INTEGER_32 is
		external "built_in"
		end

end -- class BINARY_FILE_READ
--
-- ------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is part of the SmartEiffel standard library.
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2006: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-- documentation files (the "Software"), to deal in the Software without restriction, including without
-- limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
-- the Software, and to permit persons to whom the Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
-- EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
-- AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
-- OR OTHER DEALINGS IN THE SOFTWARE.
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------
