-- See the Copyright notice at the end of this file.
--
class HTTP_SERVER

insert
	SERVER[HTTP_CONNECTION]
		export {} start, start_in_stack
		end

creation {ANY}
	stand_alone, in_stack

feature {ANY}
	is_alive: BOOLEAN is
		do
			Result := not server.done
		end

feature {HTTP_CONNECTION}
	shutdown is
			-- A connection asked the server to shut down
		do
			server.shutdown
		end

	halt is
			-- A connection asked the server to halt
		do
			server.halt
		end

	connection_done (a_connection: HTTP_CONNECTION) is
			-- A connection is about to finish.
		do
			release_connection(a_connection)
			connections := connections - 1
			if server.done then
				handle_shutdown(server)
			end
		end

feature {}
	error_handler: PROCEDURE[TUPLE[STRING]]
			-- error handler

	connection_factory: FUNCTION[TUPLE, HTTP_CONNECTION]
			-- the HTTP connection factory

	stand_alone (access: ACCESS; a_error_handler: like error_handler; a_connection_factory: like connection_factory) is
		do
			error_handler := a_error_handler
			connection_factory := a_connection_factory
			start(access)
		end

	in_stack (access: ACCESS; a_stack: like stack; a_shutdown_handler: like shutdown_handler
		a_error_handler: like error_handler; a_connection_factory: like connection_factory) is
		local
			err: BOOLEAN
		do
			error_handler := a_error_handler
			connection_factory := a_connection_factory
			err := start_in_stack(access, a_stack, a_shutdown_handler)
		end

	handle_error (msg: STRING) is
		do
			if error_handler /= Void then
				error_handler.call([msg])
			else
				std_error.put_string(msg)
				std_error.put_new_line
			end
		end

	connections_pool: FAST_ARRAY[WEAK_REFERENCE[HTTP_CONNECTION]] is
		once
			create Result.with_capacity(4)
		end

	new_connection: HTTP_CONNECTION is
		local
			i, n: INTEGER; w: WEAK_REFERENCE[HTTP_CONNECTION]
		do
			from
				i := connections_pool.lower
				n := connections_pool.upper
			until
				Result /= Void or else i > n
			loop
				w := connections_pool.item(i)
				Result := w.item
				w.set_item(Void)
				i := i + 1
			end
			if Result = Void then
				Result := connection_factory.item([])
			end
			Result.set_server(Current)
		end

	release_connection (cnx: HTTP_CONNECTION) is
		local
			i, n: INTEGER; ok: BOOLEAN; w: WEAK_REFERENCE[HTTP_CONNECTION]
		do
			from
				i := connections_pool.lower
				n := connections_pool.upper
			until
				ok or else i > n
			loop
				w := connections_pool.item(i)
				if w.item = Void then
					w.set_item(cnx)
					ok := True
				end
				i := i + 1
			end
			if not ok then
				create w.set_item(cnx)
				connections_pool.add_last(w)
			end
		end

end -- class HTTP_SERVER
--
-- ------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is part of the SmartEiffel standard library.
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2006: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-- documentation files (the "Software"), to deal in the Software without restriction, including without
-- limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
-- the Software, and to permit persons to whom the Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
-- EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
-- AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
-- OR OTHER DEALINGS IN THE SOFTWARE.
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------
