-- See the Copyright notice at the end of this file.
--
deferred class SOCKET_INPUT_OUTPUT_STREAM
	--
	-- A stream backed by a socket. You cannot directly create such a stream; instead, use ADDRESS.stream for
	-- an outbound connection; for an inbound connection, the stream is automatically created by the
	-- SOCKET_SERVER and passed to its handlers.
	--

inherit
	TERMINAL_INPUT_OUTPUT_STREAM

insert
	SOCKET_HANDLER

feature {ANY}
	is_connected: BOOLEAN is
		do
			Result := socket.is_connected
		end

	can_unread_character: BOOLEAN is
		do
			Result := index > socket.last_read.lower
		end

	end_of_input: BOOLEAN is
		do
			Result := not is_connected or else end_of_stream
		end

	disconnect is
		deferred
		end

	set_timeout (a_timeout: like timeout) is
		require
			a_timeout >= 0.0
		do
			timeout := a_timeout
		end

	unset_timeout is
			-- Makes the socket wait until some characters are available (no timeout)
		do
			timeout := no_timeout
		end

	try_again is
			-- Attempt another read. Useful when timeout is set.
		require
			is_connected
			end_of_input
		do
			ensure_read
		end

	when_disconnect (handler: PROCEDURE[TUPLE[SOCKET_INPUT_OUTPUT_STREAM]]) is
		do
			if disconnect_handlers = Void then
				create disconnect_handlers.with_capacity(2)
			end
			disconnect_handlers.add_last(handler)
		end

feature {FILTER_INPUT_STREAM}
	filtered_read_character is
		do
			index := index + 1
			ensure_read
		end

	filtered_unread_character is
		do
			index := index - 1
		end

	filtered_last_character: CHARACTER is
		do
			Result := socket.last_read.item(index)
		end

feature {FILTER_OUTPUT_STREAM}
	filtered_put_character (c: CHARACTER) is
		do
			out_buffer.add_last(c)
			if c = '%N' then
				filtered_flush
			end
		end

	filtered_flush is
		do
			socket.write(out_buffer)
			out_buffer.clear_count
		end

feature {FILTER}
	filtered_descriptor: INTEGER is
		do
			Result := socket.fd
		end

	filtered_has_descriptor: BOOLEAN is True

	filtered_stream_pointer: POINTER is
		do
			std_error.put_string("SOCKET_INPUT_OUTPUT_STREAM.filtered_stream_pointer has been called!%N")
			crash
		end

	filtered_has_stream_pointer: BOOLEAN is False

feature {}
	ensure_read is
			-- Read some new data from the socket if it is available. Does nothing if all the already read data
			-- is not yet consumed.
		do
			if socket.is_connected then
				if socket.last_read.valid_index(index) then
					end_of_stream := False
				else
					socket.read(timeout)
					end_of_stream := not socket.is_connected or else socket.last_read.is_empty
					index := 1
				end
			else
				end_of_stream := True
			end
		end

	end_of_stream: BOOLEAN

feature {}
	fire_disconnect is
		local
			i, n: INTEGER
		do
			if disconnect_handlers /= Void then
				from
					i := disconnect_handlers.lower
					n := disconnect_handlers.upper
				until
					i > n
				loop
					disconnect_handlers.item(i).call([Current])
					i := i + 1
				end
			end
		end

feature {}
	socket: SOCKET is
		deferred
		end

	make is
			-- Should be called by the creation procedures after the `socket' is set (the assertions ensure just
			-- that)
		require
			socket /= Void
			not is_made
		do
			out_buffer := ""
			unset_timeout
			socket.when_disconnected(agent socket_disconnected)
		ensure
			is_made
		end

	socket_disconnected (a_socket: SOCKET) is
		require
			a_socket = socket
		do
			fire_disconnect
		end

	made: BOOLEAN

	is_made: BOOLEAN is
			-- A trick to be sure that `made' is not alive (therefore not generated) in boost mode
		do
			Result := made
			made := True
		end

	index: INTEGER

	timeout: REAL

	out_buffer: STRING

	disconnect_handlers: FAST_ARRAY[PROCEDURE[TUPLE[SOCKET_INPUT_OUTPUT_STREAM]]]

	no_timeout: REAL is -1.0

invariant
	socket /= Void
	is_made

end -- class SOCKET_INPUT_OUTPUT_STREAM
--
-- ------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is part of the SmartEiffel standard library.
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2006: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-- documentation files (the "Software"), to deal in the Software without restriction, including without
-- limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
-- the Software, and to permit persons to whom the Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
-- EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
-- AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
-- OR OTHER DEALINGS IN THE SOFTWARE.
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------
