-- See the Copyright notice at the end of this file.
--
class ARRAY[E_]
	--
	-- General purpose resizable ARRAYs as they are define in the Eiffel language definition.
	-- The `lower' bound can be any arbitrary value, even a negative one.
	--
	-- This implementation uses only one chunk of memory, the `storage' area which is a NATIVE_ARRAY. One must
	-- keep in mind that this internal `storage' area is always kept left align. Thus, you can expect good
	-- performances while using an ARRAY to modelize a stack behavior with `add_last' / `last' / `remove_last'.
	-- Conversely `add_first' and `remove_first' are likely to slow down your program if they are too often
	-- used. If the fact that `lower' is always stuck to 0 is not a problem for you, also consider FAST_ARRAY to
	-- get better performances.
	--

inherit
	ARRAYED_COLLECTION[E_]

creation {ANY}
	make, with_capacity, from_collection, manifest_creation

feature {ANY}
	lower: INTEGER
			-- Lower index bound.

feature {ANY} -- Creation and Modification:
	make (min_index, max_index: INTEGER) is
			-- Prepare the array to hold values for indexes in range
			-- [`min_index' .. `max_index']. Set all values to default.
			-- When `max_index' = `min_index' - 1, the array `is_empty'.
		require
			valid_bounds: min_index <= max_index + 1
		do
			ensure_capacity_and_bounds(max_index - min_index + 1, min_index, max_index)
		ensure
			lower_set: lower = min_index
			upper_set: upper = max_index
			items_set: all_default
		end

	with_capacity (needed_capacity, low: INTEGER) is
			-- Create an empty array with `capacity' initialized
			-- at least to `needed_capacity' and `lower' set to `low'.
		require
			needed_capacity >= 0
		do
			ensure_capacity_and_bounds(needed_capacity, low, low - 1)
		ensure
			is_empty
			needed_capacity <= capacity
			lower = low
		end

feature {}
	ensure_capacity_and_bounds (needed_capacity, low, up: INTEGER) is
		require
			up >= low - 1
			needed_capacity >= up - low + 1
		local
			value: like item
		do
			if value = Void and then capacity > 0 then
				-- Be sure to release all objects. Useful only if the objects are references.
				clear_all
			end
			lower := low
			upper := up
			if needed_capacity > 0 then
				if capacity < needed_capacity then
					storage := storage.calloc(needed_capacity)
					capacity := needed_capacity
				elseif value /= Void and then up >= low then
					-- Ensure the postcondition. Useful only for expanded objects since references were purged above.
					clear_all
				end
			end
		ensure
			needed_capacity <= capacity
			lower = low
			upper = up
			all_default
		end

feature {ANY} -- Modification:
	resize (min_index, max_index: INTEGER) is
			-- Resize to bounds `min_index' and `max_index'. Do not lose any
			-- item whose index is in both [`lower' .. `upper'] and
			-- [`min_index' .. `max_index']. New positions if any are
			-- initialized with the appropriate default value.
		require
			min_index <= max_index + 1
		local
			needed, offset, intersize: INTEGER
		do
			needed := max_index - min_index + 1
			if needed > 0 then
				if needed > capacity then
					if capacity = 0 then
						storage := storage.calloc(needed)
						capacity := needed
					else
						storage := storage.realloc(capacity, needed)
						capacity := needed
					end
				end
				offset := lower - min_index
				intersize := max_index.min(upper) - min_index.max(lower) + 1
				if intersize > 0 then
					if offset = 0 then
						if intersize < capacity then
							storage.clear(intersize, capacity - 1)
						end
					elseif offset < 0 then
						storage.move(-offset, intersize - offset - 1, offset)
						if intersize < capacity then
							storage.clear(intersize, capacity - 1)
						end
					else
						storage.move(0, intersize - 1, offset)
						storage.clear(0, offset - 1)
						if intersize + offset < capacity then
							storage.clear(intersize + offset, capacity - 1)
						end
					end
				else
					storage.clear(0, capacity - 1)
				end
			end
			lower := min_index
			upper := max_index
		ensure
			lower = min_index
			upper = max_index
		end

	reindex (new_lower: INTEGER) is
			-- Change indexing to take in account the expected `new_lower'
			-- index. The `upper' index is translated accordingly.
		local
			i: INTEGER
		do
			i := new_lower - lower
			lower := lower + i
			upper := upper + i
		ensure
			lower = new_lower
			count = old count
		end

feature {ANY} -- Implementation of deferred:
	subarray (min, max: INTEGER): like Current is
		do
			Result := slice(min, max)
			Result.reindex(min)
		ensure then
			Result.lower = min
		end

	is_empty: BOOLEAN is
		do
			Result := upper < lower
		end

	count: INTEGER is
		do
			Result := upper - lower + 1
		end

	item (i: INTEGER): E_ is
		do
			Result := storage.item(i - lower)
		end

	put (element: like item; i: INTEGER) is
		do
			storage.put(element, i - lower)
		end

	force (element: like item; index: INTEGER) is
		require else
			True
		do
			if upper < index then
				if index = upper + 1 then
					add_last(element)
				else
					resize(lower, index)
					put(element, index)
				end
			elseif index < lower then
				resize(index, upper)
				put(element, index)
			else
				put(element, index)
			end
		ensure then
			lower = index.min(old lower)
		end

	copy (other: like Current) is
		local
			needed_capacity: INTEGER
		do
			lower := other.lower
			upper := other.upper
			needed_capacity := upper - lower + 1
			if capacity < needed_capacity then
				storage := storage.calloc(needed_capacity)
				capacity := needed_capacity
			end
			if needed_capacity > 0 then
				storage.copy_from(other.storage, needed_capacity - 1)
			end
		end

	set_all_with (v: like item) is
		do
			storage.set_all_with(v, upper - lower)
		end

	remove_first is
		do
			storage.remove_first(upper - lower)
			lower := lower + 1
		ensure then
			upper = old upper
		end

	remove_head (n: INTEGER) is
		do
			storage.move(n - lower + 1, upper - lower, -n)
			lower := lower + n
		ensure then
			upper = old upper
		end

	remove (index: INTEGER) is
		do
			storage.remove(index - lower, upper - lower)
			upper := upper - 1
		end

	clear_count, clear_count_and_capacity is
		do
			upper := lower - 1
		ensure then
			capacity = old capacity
		end

	add_first (element: like item) is
		do
			if upper < lower then
				add_last(element)
			else
				add_last(element)
				move(lower, upper - 1, 1)
				put(element, lower)
			end
		end

	add_last (element: like item) is
		local
			new_capacity: INTEGER
		do
			if capacity < count + 1 then
				if capacity = 0 then
					new_capacity := 16
					storage := storage.calloc(new_capacity)
					capacity := new_capacity
				else
					new_capacity := 2 * capacity
					storage := storage.realloc(capacity, new_capacity)
					capacity := new_capacity
				end
			end
			upper := upper + 1
			put(element, upper)
		end

	from_collection (model: COLLECTION[like item]) is
		local
			i, up: INTEGER
		do
			from
				with_capacity(model.count, model.lower)
				i := model.lower
				up := model.upper
				upper := up
			until
				i > up
			loop
				put(model.item(i), i)
				i := i + 1
			end
		ensure then
			lower = model.lower
			upper = model.upper
		end

	all_default: BOOLEAN is
		do
			Result := storage.all_default(upper - lower)
		end

	occurrences (element: like item): INTEGER is
		do
			Result := storage.occurrences(element, upper - lower)
		end

	fast_occurrences (element: like item): INTEGER is
		do
			Result := storage.fast_occurrences(element, upper - lower)
		end

	index_of (element: like item): INTEGER is
		do
			if upper >= lower then
				Result := lower + storage.first_index_of(element, upper - lower)
			else
				Result := lower
			end
		end

	reverse_index_of (element: like item; start_index: INTEGER): INTEGER is
		do
			if upper >= lower then
				Result := lower + storage.reverse_index_of(element, start_index - lower)
			else
				Result := lower
			end
		end

	fast_index_of (element: like item): INTEGER is
		do
			if upper >= lower then
				Result := lower + storage.fast_first_index_of(element, upper - lower)
			else
				Result := lower
			end
		end

	fast_reverse_index_of (element: like item; start_index: INTEGER): INTEGER is
		do
			if upper >= lower then
				Result := lower + storage.fast_reverse_index_of(element, start_index - lower)
			else
				Result := lower
			end
		end

	is_equal (other: like Current): BOOLEAN is
		do
			if Current = other then
				Result := True
			elseif lower = other.lower and then upper = other.upper then
				Result := storage.fast_memcmp(other.storage, count)
			end
		end

	is_equal_map (other: like Current): BOOLEAN is
		do
			if Current = other then
				Result := True
			elseif lower = other.lower and then upper = other.upper then
				Result := storage.memcmp(other.storage, count)
			end
		end

	slice (min, max: INTEGER): like Current is
		do
			create Result.make(lower, lower + max - min)
			if max >= min then
				-- Slice not empty
				Result.storage.slice_copy(0, storage, min - lower, max - lower)
			end
		end

	get_new_iterator: ITERATOR[E_] is
		do
			create {ITERATOR_ON_COLLECTION[E_]} Result.make(Current)
		end

feature {} -- Implement manifest generic creation.
	manifest_make (needed_capacity: INTEGER; initial_lower: INTEGER) is
			-- Manifest creation of an ARRAY[E_] with `lower' set to `initial_lower'.
		require
			needed_capacity > 0
		do
			make(initial_lower, initial_lower + needed_capacity - 1)
		end

	manifest_put (index: INTEGER; element: like item) is
		do
			check
				index < capacity
			end
			storage.put(element, index)
		end

end -- class ARRAY
--
-- ------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is part of the SmartEiffel standard library.
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2006: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-- documentation files (the "Software"), to deal in the Software without restriction, including without
-- limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
-- the Software, and to permit persons to whom the Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
-- EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
-- AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
-- OR OTHER DEALINGS IN THE SOFTWARE.
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------
