-- See the Copyright notice at the end of this file.
--
deferred class COLLECTION2[E_]
	--
	-- Abstract definition of a 2 dimensional collection of elements of type E_.
	--
	-- The SmartEiffel standard library provides two implementations of COLLECTION2: ARRAY2 and FIXED_ARRAY2.
	-- All implementations have exactly the same behavior. Switching from one implementation to
	-- another only change the memory used and the execution time.
	--

inherit
	SAFE_EQUAL[E_]
		undefine copy
		redefine fill_tagged_out_memory, is_equal
		end

feature {ANY} -- Indexing:
	lower1, lower2: INTEGER is
			-- Lower index bounds.
		deferred
		end

	frozen line_minimum: INTEGER is
			-- Equivalent of `lower1'.
		do
			Result := lower1
		end

	frozen column_minimum: INTEGER is
			-- Equivalent of `lower2'.
		do
			Result := lower2
		end

	upper1, upper2: INTEGER is
			-- Upper index bounds.
		deferred
		end

	frozen line_maximum: INTEGER is
			-- Equivalent of `upper1'.
		do
			Result := upper1
		end

	frozen column_maximum: INTEGER is
			-- Equivalent of `upper2'.
		do
			Result := upper2
		end

feature {ANY} -- Reading:
	item (line, column: INTEGER): E_ is
		require
			valid_index(line, column)
		deferred
		end

feature {ANY} -- Writing:
	put (element: like item; line, column: INTEGER) is
		require
			valid_index(line, column)
		deferred
		ensure
			item(line, column) = element
		end

	force (element: like item; line, column: INTEGER) is
			-- Put `element' at position (`line',`column'). Collection is
			-- resized first when (`line',`column') is not inside current
			-- bounds. New bounds are initialized with default values.
		require
			line >= 0
			column >= 0
		deferred
		ensure
			item(line, column) = element
			count >= old count
		end

feature {ANY} -- Index validity:
	frozen valid_line, valid_index1 (line: INTEGER): BOOLEAN is
		do
			Result := lower1 <= line and then line <= upper1
		ensure
			Result = (lower1 <= line and line <= upper1)
		end

	frozen valid_column, valid_index2 (column: INTEGER): BOOLEAN is
		do
			Result := lower2 <= column and then column <= upper2
		ensure
			Result = (lower2 <= column and column <= upper2)
		end

	frozen valid_index (line, column: INTEGER): BOOLEAN is
		do
			Result := lower1 <= line and then line <= upper1 and then lower2 <= column and then column <= upper2
		ensure
			Result = (valid_line(line) and valid_column(column))
		end

feature {ANY} -- Counting:
	count1: INTEGER is
			-- Size of the first dimension.
		deferred
		ensure
			Result = upper1 - lower1 + 1
		end

	frozen line_count: INTEGER is
			-- Equivalent of `count1'.
		do
			Result := count1
		end

	count2: INTEGER is
			-- Size of the second dimension.
		deferred
		ensure
			Result = upper2 - lower2 + 1
		end

	frozen column_count: INTEGER is
		do
			Result := count2
		end

	count: INTEGER is
			-- Total number of elements.
		deferred
		ensure
			Result = line_count * column_count
		end

feature {ANY}
	swap (line1, column1, line2, column2: INTEGER) is
			-- Swap the element at index (`line1',`column1') with the
			-- the element at index (`line2',`column2').
		require
			valid_index(line1, column1)
			valid_index(line2, column2)
		deferred
		ensure
			item(line1, column1) = old item(line2, column2)
			item(line2, column2) = old item(line1, column1)
			count = old count
		end

	set_all_with (v: like item) is
			-- Set all item with value `v'.
		deferred
		ensure
			count = old count
		end

	frozen clear_all is
			-- Set all items to default values.
		local
			value: like item
		do
			set_all_with(value)
		ensure
			count = old count
			all_default
		end

feature {ANY} -- Creating or initializing:
	from_collection2 (model: COLLECTION2[like item]) is
			--  Uses `model' to initialize Current.
		require
			model /= Void
		deferred
		ensure
			count1 = model.count1
			count2 = model.count2
		end

	from_model (model: COLLECTION[COLLECTION[E_]]) is
			-- The `model' is used to fill line by line Current.
			-- Assume all sub-collections of `model' have the same
			-- number of lines.
		require
			model.count >= 1
			model.first.count >= 1
		deferred
		ensure
			count1 = model.count
			count2 = model.first.count
		end

feature {ANY} -- Looking and comparison:
	all_default: BOOLEAN is
			-- Do all items have their type's default value?
		deferred
		end

	is_equal (other: COLLECTION2[E_]): BOOLEAN is
			-- Do both collections have the same `lower1', `lower2', `upper1' and `upper2', and items?
			-- The basic `=' is used for comparison of items.
			-- See also `is_equal_map'.
		local
			line, column: INTEGER
		do
			if lower1 /= other.lower1 then
			elseif upper1 /= other.upper1 then
			elseif lower2 /= other.lower2 then
			elseif upper2 /= other.upper2 then
			else
				from
					Result := True
					line := upper1
				until
					not Result or else line < lower1
				loop
					from
						column := upper2
					until
						not Result or else column < lower2
					loop
						Result := item(line, column) = other.item(line, column)
						column := column - 1
					end
					line := line - 1
				end
			end
		end

	is_equal_map (other: COLLECTION2[E_]): BOOLEAN is
			-- Do both collections have the same `lower1', `lower2', `upper1' and `upper2', and items?
			-- Feature `is_equal' is used for comparison of items.
			-- See also `is_equal'.
		local
			line, column: INTEGER
		do
			if lower1 /= other.lower1 then
			elseif upper1 /= other.upper1 then
			elseif lower2 /= other.lower2 then
			elseif upper2 /= other.upper2 then
			else
				from
					Result := True
					line := upper1
				until
					not Result or else line < lower1
				loop
					from
						column := upper2
					until
						not Result or else column < lower2
					loop
						Result := safe_equal(item(line, column), other.item(line, column))
						column := column - 1
					end
					line := line - 1
				end
			end
		end

	frozen same_as (other: COLLECTION2[E_]): BOOLEAN is
			-- Unlike `is_equal', this feature can be used to compare
			-- distinct implementation of COLLECTION2.
		obsolete "Now you can just use `is_equal'. (July 7th2004)"
		do
			Result := is_equal(other)
		end

feature {ANY} -- Printing:
	frozen fill_tagged_out_memory is
		local
			line, column: INTEGER; v: like item
		do
			tagged_out_memory.append(once "lower1: ")
			lower1.append_in(tagged_out_memory)
			tagged_out_memory.append(once " upper1: ")
			upper1.append_in(tagged_out_memory)
			tagged_out_memory.append(once " lower2: ")
			lower2.append_in(tagged_out_memory)
			tagged_out_memory.append(once " upper2: ")
			upper2.append_in(tagged_out_memory)
			tagged_out_memory.append(once " [%N")
			from
				line := lower1
			until
				line > upper1 or else tagged_out_memory.count > 4096
			loop
				tagged_out_memory.append(once "line ")
				line.append_in(tagged_out_memory)
				tagged_out_memory.append(once "%T: ")
				from
					column := lower2
				until
					column > upper2
				loop
					v := item(line, column)
					if v = Void then
						tagged_out_memory.append(once "Void")
					else
						v.out_in_tagged_out_memory
					end
					tagged_out_memory.extend(' ')
					column := column + 1
				end
				tagged_out_memory.extend('%N')
				line := line + 1
			end
			if valid_line(line) then
				tagged_out_memory.append(once "......%N")
			end
		end

feature {ANY} -- Miscellaneous features:
	occurrences (elt: E_): INTEGER is
			-- Number of occurrences using `is_equal'.
			-- See also `fast_occurrences' to chose the apropriate one.
		deferred
		ensure
			Result >= 0
		end

	fast_occurrences (elt: E_): INTEGER is
			-- Number of occurrences using `='.
			-- See also `occurrences' to chose the apropriate one.
		deferred
		ensure
			Result >= 0
		end

	has (x: like item): BOOLEAN is
			-- Search if a element x is in the array using `is_equal'.
			-- See also `fast_has' to chose the apropriate one.
		deferred
		end

	fast_has (x: like item): BOOLEAN is
			--  Search if a element x is in the array using `='.
		deferred
		end

	replace_all (old_value, new_value: like item) is
			-- Replace all occurrences of the element `old_value' by `new_value'
			-- using `is_equal' for comparison.
			-- See also `fast_replace_all' to choose the apropriate one.
		deferred
		ensure
			count = old count
			occurrences(old_value) = 0
		end

	fast_replace_all (old_value, new_value: like item) is
			-- Replace all occurrences of the element `old_value' by `new_value'
			-- using operator `=' for comparison.
			-- See also `replace_all' to choose the apropriate one.
		deferred
		ensure
			count = old count
			fast_occurrences(old_value) = 0
		end

	sub_collection2 (line_min, line_max, column_min, column_max: INTEGER): like Current is
			-- Create a new object using selected area of `Current'.
		require
			valid_index(line_min, column_min)
			valid_index(line_max, column_max)
		deferred
		ensure
			Result /= Void
		end

	set_area (element: like item; line_min, line_max, column_min, column_max: INTEGER) is
			-- Set all the elements of the selected area rectangle with `element'.
		require
			valid_index(line_min, column_min)
			valid_index(line_max, column_max)
		local
			line, column: INTEGER
		do
			from
				line := line_min
			until
				line > line_max
			loop
				from
					column := column_min
				until
					column > column_max
				loop
					put(element, line, column)
					column := column + 1
				end
				line := line + 1
			end
		ensure
			count = old count
		end

feature {} -- Implement manifest generic creation.
	manifest_put (index: INTEGER; element: like item) is
		require
			index >= 0
		deferred
		end

	manifest_semicolon_check: BOOLEAN is True
			-- You can use semicolons to make the end of each line more visible.

end -- class COLLECTION2
--
-- ------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is part of the SmartEiffel standard library.
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2006: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-- documentation files (the "Software"), to deal in the Software without restriction, including without
-- limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
-- the Software, and to permit persons to whom the Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
-- EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
-- AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
-- OR OTHER DEALINGS IN THE SOFTWARE.
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------
