-- See the Copyright notice at the end of this file.
--
class XML_REPOSITORY[O_ -> STORABLE]
	--
	-- A repository able to store and retrieve objects with an XML representation.
	--
	--
	-- Schema:
	--
	--   <!DOCTYPE repository [
	--
	--   <!ELEMENT repository (layout | reference | embedded | basic)* >
	--   <!ATTLIST repository version CDATA #REQUIRED >
	--
	--   <!ELEMENT layout (reference | embedded | basic | array)* >
	--   <!ATTLIST layout type CDATA #REQUIRED >
	--   <!ATTLIST layout ref ID #REQUIRED >
	--
	--   <!ELEMENT reference (EMPTY) >
	--   <!ATTLIST reference name CDARA #REQUIRED >
	--   <!ATTLIST reference ref IDREF #REQUIRED >
	--
	--   <!ELEMENT embedded (reference | embedded | basic | array)* >
	--   <!ATTLIST embedded name CDATA #REQUIRED >
	--   <!ATTLIST embedded type CDATA #REQUIRED >
	--
	--   <!ELEMENT basic (EMPTY) >
	--   <!ATTLIST basic name #REQUIRED >
	--   <!ATTLIST basic type #REQUIRED >
	--   <!ATTLIST basic value #REQUIRED >
	--
	--   <!ELEMENT array (reference | embedded | basic)* >
	--   <!ATTLIST array name CDATA #REQUIRED >
	--   <!ATTLIST array type CDATA #REQUIRED >
	--
	--   ] >
	--
	-- Vocabulary:
	--
	-- * layout: the description of the contents of an object (its layout). This object is referenced
	-- by a reference (see below).
	--
	-- * reference: a reference to an object. Note that the layout of the object MUST have been defined before
	-- any reference links to it. A special 'Void' ref indicates a Void object.
	--
	-- * embedded: a user-defined expanded object (i.e. not one of the basic objects)
	--
	-- * basic: a special expanded object, with a simple value, specially treated by the compiler and by this
	-- class. Basic types are INTEGER and co., READ and co., CHARACTER and BOOLEAN.
	--
	-- * array: a NATIVE_ARRAY of objects. The type is, in that particular case, the type of the items.
	--

inherit
	REPOSITORY[O_]
	XML_CALLBACKS

creation {ANY}
	connect_to, from_file, make

feature {ANY} -- Updating and committing
	commit_to_file (path: STRING) is
			-- Overwrite or create `path' with the contents of `Current'.
		do
			tfw.connect_to(path)
			write_to_stream(tfw)
			tfw.disconnect
		end

	update_from_file (path: STRING) is
			-- Get all objects from the file designed by `path'. Previous objects are not discarded except if new
			-- objects use the same key.
		do
			tfr.connect_to(path)
			update_from_stream(tfr)
			tfr.disconnect
		end

	read_from_file (path: STRING) is
			-- Get all objects from the file designed by `path'. Previous objects are discarded first.
		do
			tfr.connect_to(path)
			read_from_stream(tfr)
			tfr.disconnect
		end

	commit is
		do
			write_to_stream(commit_stream)
		end

	is_commitable: BOOLEAN is
		do
			Result := commit_stream /= Void and then commit_stream.is_connected
		end

	update is
		do
			read_from_stream(update_stream)
		end

	is_updateable: BOOLEAN is
		do
			Result := update_stream /= Void and then update_stream.is_connected
		end

	is_connected: BOOLEAN is
		do
			Result := commit_stream /= Void and then commit_stream.is_connected
		end

feature {ANY} -- Error handling on repository update
	register_update_error_handler (a_error_handler: PROCEDURE[TUPLE]) is
		do
			update_error_handlers.add_last(a_error_handler)
		end

	update_error_handlers: FAST_ARRAY[PROCEDURE[TUPLE]]

	fire_update_error is
		local
			i: INTEGER
		do
			sedb_breakpoint
			from
				i := update_error_handlers.lower
			until
				i > update_error_handlers.upper
			loop
				update_error_handlers.item(i).call([])
				i := i + 1
			end
		end

feature {} -- Implementation of update
	update_stream: INPUT_STREAM

	update_name: STRING is ""

	update_ref: STRING is ""

	update_type: STRING is ""

	update_value: STRING is ""

	update_capacity: STRING is ""

	update_layouts: STACK[XML_REPOSITORY_LAYOUT] is
		once
			create Result.make
		end

	open_nodes: FAST_ARRAY[STRING] is
		once
			create Result.with_capacity(4)
		end

	updated_internals: DICTIONARY[INTERNALS, STRING] is
		once
			create {HASHED_DICTIONARY[INTERNALS, STRING]} Result.make
		end

	solve (ref: STRING): INTERNALS is
		do
			Result := updated_internals.at(ref)
		end

	solver: FUNCTION[TUPLE[STRING], INTERNALS] is
		once
			Result := agent solve
		end

	xml: XML_PARSER is
		once
			create Result.make
		end

	read_from_stream (in_stream: like update_stream) is
		local
			string: STRING
		do
			repository.clear_count
			--|*** (CAD) I think it is necessary
			check
				open_nodes.is_empty
			end
			from
			until
				updated_internals.is_empty
			loop
				string := updated_internals.key(updated_internals.lower)
				updated_internals.remove(string)
				release_string(string)
			end
			update_from_stream(in_stream)
		end

	update_from_stream (in_stream: like update_stream) is
		do
			xml.connect_to(in_stream)
			xml.parse(Current)
			if not open_nodes.is_empty then
				fire_update_error
			end
		end

	record_object (ref, name: STRING) is
			-- Register the object as a high-level one, i.e. put it in the repository.
		local
			typed: TYPED_INTERNALS[O_]
		do
			if not updated_internals.has(ref) then
				fire_update_error
			else
				typed ::= solve(ref)
				put(typed.object, name.twin) --|*** memory leak, this one seems unavoidable
			end
		end

	prepare_layout (a_layout: XML_REPOSITORY_LAYOUT) is
		do
			if not update_name.is_empty then
				a_layout.set_name(update_name)
			end
			if not update_ref.is_empty then
				a_layout.set_ref(update_ref)
			end
			if not update_type.is_empty then
				a_layout.set_type(update_type)
			end
			if not update_value.is_empty then
				a_layout.set_value(update_value)
			end
			if not update_capacity.is_empty then
				a_layout.set_capacity(update_capacity.to_integer)
			end
		end

feature {XML_PARSER}
	with_attribute (attribute_name: STRING; attribute_value: STRING) is
		do
			inspect
				attribute_name
			when "name" then
				update_name.copy(attribute_value)
			when "type" then
				update_type.copy(attribute_value)
			when "ref" then
				update_ref.copy(attribute_value)
			when "value" then
				update_value.copy(attribute_value)
			when "capacity" then
				update_capacity.copy(attribute_value)
			when "version" then
				if not version.is_equal(attribute_value) then
					fire_update_error
				end
			else
				fire_update_error
			end
		end

	open_node (node_name: STRING) is
		local
			s: STRING; layout: XML_REPOSITORY_LAYOUT
		do
			s := new_string
			s.copy(node_name)
			open_nodes.add_last(s)
			layout := new_layout
			layout.set(node_name)
			update_layouts.push(layout)
			inspect
				node_name
			when "repository" then
				-- nothing
			when "layout" then
				prepare_layout(layout)
			when "reference" then
				prepare_layout(layout)
			when "embedded" then
				prepare_layout(layout)
			when "basic" then
				prepare_layout(layout)
			when "array" then
				prepare_layout(layout)
			else
				fire_update_error
			end
			update_name.clear_count
			update_ref.clear_count
			update_type.clear_count
			update_value.clear_count
			update_capacity.clear_count
		end

	close_node (node_name: STRING) is
		local
			layout: XML_REPOSITORY_LAYOUT; internals: INTERNALS; ref: STRING
		do
			release_string(open_nodes.last)
			open_nodes.remove_last
			layout := update_layouts.top
			update_layouts.pop
			inspect
				node_name
			when "repository" then
				-- it's all over
			when "layout" then
				internals := layout.solve(solver)
				internals.set_object_can_be_retrieved
				if updated_internals.has(layout.ref) then
					ref := updated_internals.key(index_of_key(layout.ref))
					updated_internals.remove(ref)
				else
					ref := new_string
				end
				ref.copy(layout.ref)
				updated_internals.add(internals, ref)
			when "reference" then
				if update_layouts.count = 1 then
					-- means only the "repository" node is open
					record_object(layout.ref, layout.name)
				else
					update_layouts.top.add_layout(layout)
				end
			when "embedded" then
				update_layouts.top.add_layout(layout)
			when "basic" then
				update_layouts.top.add_layout(layout)
			when "array" then
				update_layouts.top.add_layout(layout)
			else
				fire_update_error
			end
		end

	open_close_node (node_name: STRING) is
		do
			open_node(node_name)
			close_node(node_name)
		end

	current_node: STRING is
		do
			Result := open_nodes.last
		end

	data (a_data: STRING) is
		local
			i: INTEGER; break: BOOLEAN
		do
			from
				i := a_data.lower
			until
				i > a_data.upper or else break
			loop
				if not a_data.item(i).is_separator then
					fire_update_error
				end
				i := i + 1
			end
		end

	parse_error is
		do
			fire_update_error
		end

feature {} -- Implementation of commit
	commit_stream: OUTPUT_STREAM

	commit_map: SET[POINTER] is
			-- Used when committing object not to commit them twice
		once
			create {HASHED_SET[POINTER]} Result.make
		end

	write_to_stream (out_stream: like commit_stream) is
		require
			out_stream.is_connected
		local
			i: INTEGER
		do
			commit_map.clear_count
			out_stream.put_string(once "<repository version='")
			out_stream.put_string(version)
			out_stream.put_string(once "'>%N")
			from
				i := lower
			until
				i > upper
			loop
				write_object(key(i), item(i), out_stream)
				i := i + 1
			end
			out_stream.put_string(once "</repository>%N")
		end

	write_object (name: like key; object: like item; out_stream: like commit_stream) is
		local
			int: INTERNALS
		do
			if object = Void then
				write_reference(Void, name, out_stream)
			else
				int := object.to_internals
				if int.type_is_expanded then
					write_expanded(int, name, out_stream)
				else
					if not commit_map.has(int.object_as_pointer) then
						write_layout(int, out_stream)
					end
					write_reference(int, name, out_stream)
				end
			end
		end

	write_reference (layout: INTERNALS; name: STRING; out_stream: like commit_stream) is
		require
			not layout.type_is_expanded
		do
			out_stream.put_string(once "<reference name='")
			out_stream.put_string(name)
			if layout = Void then
				out_stream.put_string(once "' ref='Void'")
			else
				out_stream.put_string(once "' ref='")
				out_stream.put_integer(layout.object_as_pointer.hash_code)
				out_stream.put_string(once "'/>%N")
			end
		end

	write_layout (layout: INTERNALS; out_stream: like commit_stream) is
		require
			not commit_map.has(layout.object_as_pointer)
			not layout.type_is_expanded
		local
			i: INTEGER; int: INTERNALS
		do
			-- Add the pointer to the map of "known objects" (those already written). It must be done first
			-- because of possible recursion
			commit_map.add(layout.object_as_pointer)
			-- Write the nested objects not already defined
			from
				i := 1
			until
				i > layout.type_attribute_count
			loop
				int := layout.object_attribute(i)
				if int /= Void then
					if int.type_is_expanded then
						if int.type_is_native_array then
							write_array_fields_layouts(int, out_stream)
						end
					elseif not commit_map.has(int.object_as_pointer) then
						write_layout(int, out_stream)
					end
				end
				i := i + 1
			end
			out_stream.put_string(once "<layout ref='")
			out_stream.put_integer(layout.object_as_pointer.hash_code)
			out_stream.put_string(once "' type='")
			out_stream.put_string(layout.type_generating_type)
			out_stream.put_string(once "'>%N")
			write_contents(layout, out_stream)
			out_stream.put_string(once "</layout>%N")
		ensure
			commit_map.has(layout.object_as_pointer)
		end

	write_contents (layout: INTERNALS; out_stream: like commit_stream) is
		local
			i: INTEGER; int: INTERNALS
		do
			from
				i := 1
			until
				i > layout.type_attribute_count
			loop
				int := layout.object_attribute(i)
				if int = Void then
					write_reference(Void, layout.type_attribute_name(i), out_stream)
				else
					if int.type_is_expanded then
						write_expanded(int, layout.type_attribute_name(i), out_stream)
					else
						write_reference(int, layout.type_attribute_name(i), out_stream)
					end
				end
				i := i + 1
			end
		end

	write_array_fields_layouts (array: INTERNALS; out_stream: like commit_stream) is
		require
			array.type_is_expanded and then array.type_is_native_array
		local
			f: INTEGER; int: INTERNALS
		do
			from
				f := 1
			until
				f > array.type_attribute_count
			loop
				int := array.object_attribute(f)
				if int /= Void and then not int.type_is_expanded and then not commit_map.has(int.object_as_pointer) then
					write_layout(int, out_stream)
				end
				f := f + 1
			end
		end

	write_expanded (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		require
			internals.type_is_expanded
		local
			type: STRING
		do
			type := internals.type_generating_type
			inspect
				type
			when "CHARACTER" then
				write_character_layout_object(internals, name, out_stream)
			when "BOOLEAN" then
				write_boolean_layout_object(internals, name, out_stream)
			when "INTEGER_8" then
				write_integer_8_layout_object(internals, name, out_stream)
			when "INTEGER_16" then
				write_integer_16_layout_object(internals, name, out_stream)
			when "INTEGER" then
				write_integer_layout_object(internals, name, out_stream)
			when "INTEGER_32" then
				write_integer_32_layout_object(internals, name, out_stream)
			when "INTEGER_64" then
				write_integer_64_layout_object(internals, name, out_stream)
			when "REAL" then
				write_real_layout_object(internals, name, out_stream)
			when "REAL_32" then
				write_real_32_layout_object(internals, name, out_stream)
			when "REAL_64" then
				write_real_64_layout_object(internals, name, out_stream)
			when "REAL_80" then
				write_real_80_layout_object(internals, name, out_stream)
			when "REAL_128" then
				write_real_128_layout_object(internals, name, out_stream)
			when "REAL_EXTENDED" then
				write_real_expanded_layout_object(internals, name, out_stream)
			else
				if internals.type_is_native_array then
					write_array_layout_object(internals, name, out_stream)
				else
					write_embedded_layout_object(internals, name, out_stream)
				end
			end
		end

	write_character_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[CHARACTER]; c: CHARACTER
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='CHARACTER' value='")
			out_stream.put_character(c)
			out_stream.put_string(once "'/>%N")
		end

	write_boolean_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[BOOLEAN]; c: BOOLEAN
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='BOOLEAN' value='")
			out_stream.put_boolean(c)
			out_stream.put_string(once "'/>%N")
		end

	write_integer_8_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[INTEGER_8]; c: INTEGER_8
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='INTEGER_8' value='")
			out_stream.put_integer(c)
			out_stream.put_string(once "'/>%N")
		end

	write_integer_16_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[INTEGER_16]; c: INTEGER_16
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='INTEGER_16' value='")
			out_stream.put_integer(c)
			out_stream.put_string(once "'/>%N")
		end

	write_integer_32_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[INTEGER_32]; c: INTEGER_32
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='INTEGER_32' value='")
			out_stream.put_integer(c)
			out_stream.put_string(once "'/>%N")
		end

	write_integer_64_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[INTEGER_64]; c: INTEGER_64
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='INTEGER_64' value='")
			out_stream.put_integer(c)
			out_stream.put_string(once "'/>%N")
		end

	write_integer_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[INTEGER]; c: INTEGER
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='INTEGER' value='")
			out_stream.put_integer(c)
			out_stream.put_string(once "'/>%N")
		end

	write_real_32_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[REAL_32]; c: REAL_32
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='REAL_32' value='")
			out_stream.put_real(c)
			out_stream.put_string(once "'/>%N")
		end

	write_real_64_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[REAL_64]; c: REAL_64
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='REAL_64' value='")
			out_stream.put_real(c)
			out_stream.put_string(once "'/>%N")
		end

	write_real_80_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[REAL_80]; c: REAL_80
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='REAL_80' value='")
			out_stream.put_real(c.force_to_real_64)
			out_stream.put_string(once "'/>%N")
		end

	write_real_128_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[REAL_128]; c: REAL_128
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='REAL_128' value='")
			out_stream.put_real(c.force_to_real_64)
			out_stream.put_string(once "'/>%N")
		end

	write_real_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[REAL]; c: REAL
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='REAL' value='")
			out_stream.put_real(c)
			out_stream.put_string(once "'/>%N")
		end

	write_real_expanded_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			t: TYPED_INTERNALS[REAL_EXTENDED]; c: REAL_EXTENDED
		do
			t ::= internals
			c := t.object
			out_stream.put_string(once "<basic name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='REAL_EXTENDED' value='")
			out_stream.put_real(c.force_to_real_64)
			out_stream.put_string(once "'/>%N")
		end

	write_array_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		local
			type: STRING; f: INTEGER; int: INTERNALS
		do
			type := once ""
			type.copy(internals.type_generating_type)
			type.remove_prefix(once "NATIVE_ARRAY[")
			type.remove_suffix(once "]")
			out_stream.put_string(once "<array name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='")
			out_stream.put_string(type)
			out_stream.put_string(once "' capacity='")
			out_stream.put_integer(internals.type_attribute_count)
			out_stream.put_string(once "'>%N")
			from
				f := 1
			until
				f > internals.type_attribute_count
			loop
				int := internals.object_attribute(f)
				if int = Void then
					write_reference(Void, internals.type_attribute_name(f), out_stream)
				else
					if int.type_is_expanded then
						write_expanded(int, internals.type_attribute_name(f), out_stream)
					else
						write_reference(int, internals.type_attribute_name(f), out_stream)
					end
				end
				f := f + 1
			end
			out_stream.put_string(once "</array>%N")
		end

	write_embedded_layout_object (internals: INTERNALS; name: STRING; out_stream: like commit_stream) is
		do
			out_stream.put_string(once "<embedded name='")
			out_stream.put_string(name)
			out_stream.put_string(once "' type='")
			out_stream.put_string(internals.type_generating_type)
			out_stream.put_string(once "'>%N")
			write_contents(internals, out_stream)
			out_stream.put_string(once "</embedded>%N")
		end

feature {} -- Creation
	from_file (path: STRING) is
			-- Create a not-connected repository with initial data from the given path. One must use
			-- `update_from_file' and `commit_to_file'.
		do
			make
			update_from_file(path)
		end

	make is
			-- Create a not-connected empty repository. One must use `update_from_file' and `commit_to_file'.
		do
			create {HASHED_DICTIONARY[O_, STRING]} repository.make
			create update_error_handlers.with_capacity(2)
		end

	connect_to (in_stream: like update_stream; out_stream: like commit_stream) is
			-- Connect to a repository with streams as physical store.
		do
			make
			update_stream := in_stream
			commit_stream := out_stream
		end

feature {} -- Miscellany
	tfr: TEXT_FILE_READ is
		once
			create Result.make
		end

	tfw: TEXT_FILE_WRITE is
		once
			create Result.make
		end

	frozen version: STRING is "1"

	free_layouts: STACK[XML_REPOSITORY_LAYOUT] is
		once
			create Result.make
		end

	new_layout: XML_REPOSITORY_LAYOUT is
		do
			if free_layouts.is_empty then
				create Result.make
			else
				Result := free_layouts.top
				free_layouts.pop
			end
		ensure
			Result.is_clear
		end

	release_layout (a_layout: XML_REPOSITORY_LAYOUT) is
		do
			a_layout.clear
			free_layouts.push(a_layout)
		end

	free_strings: STACK[STRING] is
		once
			create Result.make
		end

	new_string: STRING is
		do
			if free_strings.is_empty then
				Result := ""
			else
				Result := free_strings.top
				free_strings.pop
			end
		ensure
			Result.is_empty
		end

	release_string (a_string: STRING) is
		do
			a_string.clear_count
			free_strings.push(a_string)
		end

	index_of_key (a_string: STRING): INTEGER is
		local
			i: INTEGER; done: BOOLEAN
		do
			from
				i := updated_internals.lower
			until
				done
			loop
				if updated_internals.key(i).is_equal(a_string) then
					Result := i
					done := True
				elseif i = updated_internals.upper then
					done := True
				else
					i := i + 1
				end
			end
		end

end -- class XML_REPOSITORY
--
-- ------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- This file is part of the SmartEiffel standard library.
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2006: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
-- documentation files (the "Software"), to deal in the Software without restriction, including without
-- limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
-- the Software, and to permit persons to whom the Software is furnished to do so, subject to the following
-- conditions:
--
-- The above copyright notice and this permission notice shall be included in all copies or substantial
-- portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
-- LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
-- EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
-- AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
-- OR OTHER DEALINGS IN THE SOFTWARE.
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------
