-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class EIFFELDOC
	--
	-- The html documentation tool. It generates a whole site that describes a project, given a loadpath.se
	-- file or an ACE file.
	--
	-- If no such file is provided, the classes from the configuration file are used.
	--

inherit
	EXTERNAL_TOOL
		export {EIFFELDOC_OPTIONS} flag_match
		end
	VISITOR
	CLUSTER_VISITOR

creation {ANY}
	make

feature {ANY}
	usage: STRING is
		once
			Result := "Usage: "
			Result.append(command_name)
			Result.extend(' ')
			Result.append(options.command_usage)
			Result.append("{-prune cluster...} {-verbose} {-help} [loadpath.se | ACE file]%N")
			Result.extend('%N')
			Result.append(options.command_help)
			Result.append("[

								-prune    helps avoiding the clusters you don't want to eiffeldoc. It
								          uses the loadpath.se or ACE syntax for paths.
								          Note that it also removes all the subclusters.
								          For example:


								]")
			Result.append("          ")
			Result.append(command_name)
			Result.append(" -prune 'lib/scoop'%N")
			Result.append("[

								The loadpath.se or the ACE file is not compulsory. When not set, the
								classes of the current directory are considered, and the classes of
								the default loadpath.

								Note that you should use a loadpath.se or an ACE file for more sensible
								results. -prune should be avoided whenever possible (it usually leads to
								inconsistent HTML pages, with dangling links). Use at your own risk.

								]")
		end

feature {}
	make is
		local
			o: like options
		do
			create pruned_clusters.make(0)
			create cluster_html.make
			create all_clusters.make(0)
			create clusters.make(0)
			create classes.make
			create all_classes.make(0)
			create comments.make
			create o.make(Current)
			context.set_options(o)

			bootstrap
			options.set_default_arguments
			get_started
			write_index
			options.generate_stylesheets
		end

	get_started is
		local
			i: INTEGER; c: CLUSTER
		do
			create core.make(context)
			from
				i := 0
			until
				i = ace.cluster_count
			loop
				c := ace.cluster_at(i)
				echo.put_string(once "Loading cluster %"")
				echo.put_string(c.name)
				echo.put_string(once "%"%N")
				c.accept(Current)
				i := i + 1
			end
			core.do_doc
		end

	context: EIFFELDOC_CONTEXT is
		once
			create Result.make
		end

	options: EIFFELDOC_OPTIONS is
		do
			Result := context.options
		end

	core: EIFFELDOC_CORE

	parse_arguments is
		local
			i, n: INTEGER
		do
			search_for_verbose_flag
			root_class_name := once "NONE"
			from
				i := 1
			until
				i > argument_count
			loop
				if flag_match(once "prune", argument(i)) then
					-- prune this cluster
					if i + 1 > argument_count then
						std_error.put_string(usage)
						die_with_code(1)
					end
					prune_cluster(argument(i + 1))
					i := i + 2
				elseif is_help_flag(argument(i)) then
					i := i + 1
				elseif is_version_flag(argument(i)) then
					i := i + 1
				elseif is_verbose_flag(argument(i)) then
					i := i + 1
				else
					n := options.parse_argument(i)
					if n > 0 then
						debug
							echo.put_string(once "Argument #")
							echo.put_integer(i)
							echo.put_string(once ": ")
							echo.put_string(argument(i))
							echo.put_string(once " spans ")
							if n = 1 then
								echo.put_string(once "1 argument%N")
							else
								echo.put_integer(n)
								echo.put_string(once " arguments%N")
							end
						end

						i := i + n
					else
						if i = argument_count and then set_loadpath(argument(i)) then
							i := i + 1
						else
							std_error.put_string(usage)
							die_with_code(1)
						end
					end
				end
			end
		end

	set_loadpath (loadpath: STRING): BOOLEAN is
		do
			if loadpath.has_suffix(once ".ace") or else loadpath.has_suffix(once ".ACE") then
				-- nothing to do: bootstrap takes care of that
				Result := True
			elseif loadpath.has_suffix(once ".se") or else loadpath.has_suffix(once ".SE") then
				system_tools.add_loadpath_file(loadpath)
				Result := True
			end
		end

	use_short_mode: BOOLEAN is True

	is_valid_argument_for_ace_mode (arg: STRING): BOOLEAN is
		do
			-- always False
		end

	valid_argument_for_ace_mode: STRING is ""

	is_valid_class_name (name: STRING): BOOLEAN is
			-- Does it follow the syntactic rules for a class name?  I.e. an uppercase letter followed by
			-- optional uppercase letters, digits and underscores.
		local
			i, n: INTEGER
		do
			if not name.is_empty then
				n := name.count
				from
					i := 2
					inspect name.first
					when 'A'..'Z' then
						Result := True
					end
				until
					not Result or else i > n
				loop
					inspect name.item(i)
					when 'A'..'Z', '0'..'9', '_' then
					else
						Result := False
					end
					i := i + 1
				end
			end
		end

feature {CLUSTER}
	visit_cluster (visited: CLUSTER) is
		local
			bd: BASIC_DIRECTORY; p, s: STRING; cn: HASHED_STRING; ct: CLASS_TEXT
		do
			if pruned_cluster(visited) then
				echo.put_string(once "Cluster ")
				echo.put_string(cluster_name(visited))
				echo.put_string(once " pruned%N")
			else
				echo.put_string(once "Cluster ")
				echo.put_string(cluster_name(visited))
				echo.put_string(once " included%N")

				s := visited.directory_path
				if s /= Void and then not s.is_empty then
					bd.connect_to(s)
					if bd.is_connected then
						from
							s := once ""
							p := once ""
							bd.read_entry
						until
							bd.end_of_input
						loop
							s.copy(bd.last_entry)
							p.copy(visited.directory_path)
							bd.compute_file_path_with(p, s)
							p.copy(bd.last_entry)
							-- p is the complete path of the file
							s.to_upper
							if s.has_suffix(once ".E") then
								s.remove_tail(2)
								-- s is now simply the class name
								cn := string_aliaser.hashed_string(s)
								if is_valid_class_name(s) then
									ct := smart_eiffel.class_text(create {CLASS_NAME}.unknown_position(cn), True)
									if ct.cluster = visited then
										-- how could it be otherwise?
										add_class(ct)
									end
								end
							end
							bd.read_entry
						end
						bd.disconnect
					end
				end
				write_cluster(visited)
			end
		end

feature {} -- Pruned clusters
	prune_cluster (cluster: STRING) is
		local
			cluster_tree: STRING
		do
			cluster_tree := cluster.twin
			pruned_clusters.add_last(cluster_tree)
		end

	pruned_cluster (cluster: CLUSTER): BOOLEAN is
		local
			i: INTEGER; name, s: STRING
		do
			name := cluster_name(cluster)
			from
				i := pruned_clusters.lower
			until
				Result or else i > pruned_clusters.upper
			loop
				s := pruned_clusters.item(i)
				Result := name.same_as(s)
				if not Result then
					s := once ""
					s.copy(pruned_clusters.item(i))
					s.extend('/')
					Result := name.has_prefix(s)
				end
				i := i + 1
			end
		end

	pruned_clusters: FAST_ARRAY[STRING]

feature {} -- General HTML files
	write_index is
		local
			f: TEXT_FILE_WRITE; html: HTML_OUTPUT_STREAM
		once
			create f.connect_to("index.html")
			create html.connect_to(f)
			html.header
			html.open_title
			html.put_string(options.title)
			html.close_title
			html.frames
			html.with_attribute("cols", "250px,*")
			html.with_attribute("border", "0")
			html.with_attribute("frameborder", "no")
			html.open_frameset
			if options.home_address /= Void then
				html.with_attribute("rows", "56px,30%%,5px,*")
			else
				html.with_attribute("rows", "30%%,5px,*")
			end
			html.with_attribute("border", "0")
			html.with_attribute("frameborder", "no")
			html.with_attribute("framespacing", "0")
			html.open_frameset
			if options.home_address /= Void then
				html.with_attribute("frameborder", "no")
				html.with_attribute("scrolling", "no")
				html.put_frame("_home.html", Void)
			end
			html.with_attribute("frameborder", "no")
			html.put_frame("_clusters.html", Void)
			html.with_attribute("frameborder", "no")
			html.with_attribute("scrolling", "no")
			html.put_frame("_separator.html", Void)
			html.with_attribute("frameborder", "no")
			html.put_frame("_classes.html", "classes")
			html.close_frameset
			html.put_frame("_summary.html", "doc")
			html.close_frameset
			html.open_noframes
			html.open_subsection
			html.put_string(once "Frame Alert")
			html.close_subsection
			html.open_paragraph
			html.put_string(once "This document is designed to be viewed exclusively using the frames feature.")
			html.close_paragraph
			html.close_noframes
			html.close
			html.disconnect
			if options.home_address /= Void then
				write_home
			end
			write_clusters
			write_separator
			write_classes
			write_summaries
		end

	write_home is
		require
			options.home_address /= Void
		local
			html: HTML_OUTPUT_STREAM
		once
			html := options.navbar_html("_home.html")
			html.with_attribute("border", "0")
			html.with_attribute("cellspacing", "3")
			html.with_attribute("cellpadding", "3")
			html.with_attribute("cols", "1")
			html.with_attribute("width", "220")
			html.open_table
			html.open_table_row
			html.with_attribute("align", "center")
			html.with_attribute("class", "home")
			html.open_table_cell
			html.with_attribute("class", "home")
			html.open_anchor_address(options.home_address, "_top")
			if options.home_image_address /= Void then
				html.with_attribute("alt", "Home")
				html.with_attribute("border", "0")
				html.put_image(options.home_image_address)
			else
				html.put_string("Home")
			end
			html.close_anchor
			html.close_table_cell
			html.close_table_row
			html.close_table
			html.close
			html.disconnect
		end

	write_clusters is
		local
			i: INTEGER; s: STRING; html: HTML_OUTPUT_STREAM; cluster: STRING
		once
			html := options.navbar_html("_clusters.html")
			html.with_attribute("border", "0")
			html.with_attribute("cellspacing", "3")
			html.with_attribute("cellpadding", "3")
			html.with_attribute("cols", "1")
			html.with_attribute("width", "100%%")
			html.open_table
			html.open_table_row
			html.with_attribute("nowrap", Void)
			html.with_attribute("class", "cluster")
			html.open_table_cell
			html.with_attribute("class", "cluster")
			html.open_anchor_address("_classes.html", "classes")
			put_arrow(html)
			html.put_string("All clusters")
			html.close_anchor
			html.put_break
			from
				s := ""
				i := clusters.lower
			until
				i > clusters.upper
			loop
				cluster := clusters.item(i)
				s.copy(once "_classes_")
				s.append(cluster_pathname(cluster))
				s.append(once ".html")
				html.with_attribute(once "class", once "cluster")
				html.open_anchor_address(s, "classes")
				put_arrow(html)
				html.put_string(cluster)
				html.close_anchor
				html.put_break
				i := i + 1
			end
			html.close_table_cell
			html.close_table_row
			html.close_table
			html.close
			html.disconnect
		end

	write_separator is
		local
			html: HTML_OUTPUT_STREAM
		once
			html := options.navbar_html("_separator.html")
			html.put_entity("nbsp")
			html.close
			html.disconnect
		end

	write_classes is
		local
			i: INTEGER; cluster, filename, s: STRING
		once
			write_classes_for("All clusters", "_classes.html", "_summary.html", all_classes)
			from
				s := ""
				filename := ""
				i := clusters.lower
			until
				i > clusters.upper
			loop
				cluster := clusters.item(i)
				filename.copy(once "_classes_")
				filename.append(cluster_pathname(cluster))
				filename.append(once ".html")
				s.copy(once "_summary@")
				s.append(cluster_pathname(cluster))
				s.append(once ".html")
				write_classes_for(cluster, filename, s, classes.reference_at(cluster_name(all_clusters.item(i))))
				i := i + 1
			end
		end

	write_classes_for (cluster, filename, summary_link: STRING; list: FAST_ARRAY[STRING]) is
		require
			cluster /= Void
		local
			i: INTEGER; s: STRING; html: HTML_OUTPUT_STREAM; clazz: STRING
		do
			html := options.navbar_html(filename)
			html.with_attribute(once "border", once "0")
			html.with_attribute(once "width", once "100%%")
			html.open_table
			html.open_table_row
			html.with_attribute(once "align", once "center")
			html.with_attribute(once "valign", once "center")
			html.with_attribute(once "class", once "title")
			html.open_table_cell
			html.with_attribute(once "class", once "title")
			html.open_anchor_address(summary_link, once "doc")
			html.put_string(cluster)
			html.close_anchor
			html.close_table_cell
			html.close_table_row
			html.close_table
			if list /= Void then
				html.put_break
				html.with_attribute(once "border", once "0")
				html.with_attribute(once "cellspacing", once "3")
				html.with_attribute(once "cellpadding", once "3")
				html.with_attribute(once "width", once "100%%")
				html.open_table
				html.open_table_row
				html.with_attribute(once "class", once "class")
				html.with_attribute(once "nowrap", Void)
				html.open_table_cell
				from
					s := once ""
					i := list.lower
				until
					i > list.upper
				loop
					clazz := list.item(i)
					s.copy(once "_class@")
					s.append(clazz)
					s.append(once ".html")
					html.with_attribute(once "class", once "class")
					html.open_anchor_address(s, once "doc")
					put_arrow(html)
					html.put_string(clazz)
					html.close_anchor
					html.put_break
					i := i + 1
				end
				html.close_table_cell
				html.close_table_row
				html.close_table
			end
			html.close
			html.disconnect
		end

	write_summaries is
		local
			i: INTEGER; cluster: STRING
		once
			write_summary
			from
				i := clusters.lower
			until
				i > clusters.upper
			loop
				cluster := clusters.item(i)
				debug
					echo.put_string(once "Writing summary for ")
					echo.put_string(cluster)
					echo.put_new_line
				end
				write_cluster_summary(cluster)
				i := i + 1
			end
		end

	write_summary is
		local
			i: INTEGER; s: STRING; html: HTML_OUTPUT_STREAM; cluster: STRING; html_parser: HTML_PARSER
		once
			html := options.doc_html("_summary.html")
			html.body
			html.with_attribute("width", "100%%")
			html.with_attribute("border", "0")
			html.open_table
			html.open_table_row
			html.with_attribute("valign", "center")
			html.with_attribute("align", "center")
			html.with_attribute("class", "title")
			html.open_table_cell
			html.put_string("All clusters")
			html.close_table_cell
			html.close_table_row
			html.close_table
			html.put_break
			html.with_attribute("border", "0")
			html.with_attribute("width", "100%%")
			html.open_table
			from
				s := ""
				i := clusters.lower
			until
				i > clusters.upper
			loop
				cluster := clusters.item(i)
				s.copy(once "_summary@")
				s.append(cluster_pathname(cluster))
				s.append(once ".html")
				html.open_table_row
				html.with_attribute(once "valign", once "top")
				html.with_attribute(once "class", once "cluster")
				html.open_table_cell
				html.with_attribute(once "class", once "cluster")
				html.open_anchor_address(s, Void)
				html.put_string(cluster)
				html.close_anchor
				html.close_table_cell
				html.with_attribute(once "align", once "justify")
				html.open_table_cell
				html_parser := cluster_html.reference_at(cluster)
				if html_parser /= Void then
					write_description(html_parser, html, True)
				end
				html.put_break
				html.put_entity(once "nbsp")
				html.close_table_cell
				html.close_table_row
				i := i + 1
			end
			html.close_table
			html.close
			html.disconnect
		end

	write_cluster_summary (a_cluster: STRING) is
		local
			i: INTEGER; s: STRING; html: HTML_OUTPUT_STREAM; clazz: STRING; html_parser: HTML_PARSER
			cluster_classes: FAST_ARRAY[STRING]; comment: COMMENT
		do
			cluster_classes := classes.reference_at(a_cluster)
			if cluster_classes /= Void then
				s := once ""
				s.copy(once "_summary@")
				s.append(cluster_pathname(a_cluster))
				s.append(once ".html")
				html := options.doc_html(s)
				html.body
				html_parser := cluster_html.reference_at(a_cluster)

				html.with_attribute(once "width", once "100%%")
				html.with_attribute(once "border", once "0")
				html.open_table
				html.open_table_row
				html.with_attribute(once "valign", once "center")
				html.with_attribute(once "align", once "center")
				html.with_attribute(once "class", once "title")
				html.open_table_cell
				html.put_string(once "Cluster ")
				html.put_string(a_cluster)
				html.close_table_cell
				html.close_table_row
				html.close_table
				html.put_break
				if html_parser /= Void then
					html.with_attribute(once "width", once "100%%")
					html.with_attribute(once "border", once "0")
					html.with_attribute(once "cellspacing", once "0")
					html.open_table
					html.open_table_row
					html.with_attribute(once "width", once "60%%")
					html.open_table_cell
					html.put_entity(once "nbsp")
					html.close_table_cell
					html.with_attribute(once "valign", once "top")
					html.with_attribute(once "align", once "left")
					html.with_attribute(once "class", once "description")
					html.open_table_cell
					write_description(html_parser, html, True)
					html.close_table_cell
					html.close_table_row
					html.open_table_row
					html.with_attribute(once "width", once "60%%")
					html.open_table_cell
					html.put_entity(once "nbsp")
					html.close_table_cell
					html.with_attribute(once "valign", once "bottom")
					html.with_attribute(once "align", once "right")
					html.with_attribute(once "class", once "description")
					html.open_table_cell
					html.with_attribute(once "class", once "description")
					html.open_anchor_address(once "#desc", Void)
					put_arrow(html)
					html.put_string("See description")
					html.close_anchor
					html.close_table_cell
					html.close_table_row
					html.close_table
					html.put_break
				end
				html.with_attribute(once "border", once "0")
				html.with_attribute(once "width", once "100%%")
				html.open_table
				from
					s := ""
					i := cluster_classes.lower
				until
					i > cluster_classes.upper
				loop
					clazz := cluster_classes.item(i)
					s.copy(once "_class@")
					s.append(clazz)
					s.append(once ".html")
					html.open_table_row
					html.with_attribute(once "valign", once "top")
					html.with_attribute(once "class", once "class")
					html.open_table_cell
					html.with_attribute(once "class", once "class")
					html.open_anchor_address(s, Void)
					html.put_string(clazz)
					html.close_anchor
					html.close_table_cell
					html.with_attribute(once "align", once "justify")
					html.open_table_cell
					comment := comments.reference_at(clazz)
					if comment /= Void then
						context.set_html(html)
						context.set_class_text(smart_eiffel.class_text(create {CLASS_NAME}.unknown_position(string_aliaser.hashed_string(clazz)), True))
						context.write_comment(comment, True)
					end
					html.put_break
					html.put_entity(once "nbsp")
					html.close_table_cell
					html.close_table_row
					i := i + 1
				end
				html.close_table
				if html_parser /= Void then
					put_separator(html)
					html.open_anchor_name(once "desc")
					html.open_section
					html.put_string(once "Description")
					html.close_section
					html.close_anchor
					write_description(html_parser, html, False)
				end
				html.close
				html.disconnect
			end
		end

	write_cluster (cluster: CLUSTER) is
		do
			add_cluster(cluster)
		end

feature {} -- Support features for html writing
	all_clusters: FAST_ARRAY[CLUSTER]

	clusters: FAST_ARRAY[STRING]

	cluster_html: HASHED_DICTIONARY[HTML_PARSER, STRING]

	classes: HASHED_DICTIONARY[FAST_ARRAY[STRING], STRING]

	all_classes: FAST_ARRAY[STRING]

	comments: HASHED_DICTIONARY[COMMENT, STRING]

	cluster_summary: TEXT_FILE_READ is
		once
			create Result.make
		end

	body_node: EIFFELDOC_HTML is
		once
			create Result.make
		end

	cluster_name (a_cluster: CLUSTER): STRING is
		do
			Result := a_cluster.name
		end

	cluster_pathname (a_cluster: STRING): STRING is
		do
			Result := once ""
			Result.copy(a_cluster)
			Result.replace_all('/', '.')
		end

	add_cluster (cluster: CLUSTER) is
		local
			i: INTEGER; s: STRING; sorter: COLLECTION_SORTER[STRING]; bd: BASIC_DIRECTORY; ft: FILE_TOOLS
			html_parser: HTML_PARSER
		do
			s := cluster_name(cluster).twin
			i := sorter.insert_index(clusters, s)
			clusters.add(s, i)
			all_clusters.add(cluster, i)
			bd.compute_file_path_with(cluster.directory_path, once "cluster.html")
			debug
				io.put_string(bd.last_entry)
				io.put_new_line
			end
			s := once ""
			s.copy(bd.last_entry)
			if ft.is_readable(s) then
				debug
					io.put_string(once " (readable)%N")
				end
				cluster_summary.connect_to(s)
				create html_parser.parse(cluster_summary)
				cluster_summary.disconnect
				debug
					if body(html_parser) /= Void then
						io.put_string(body(html_parser).to_string)
					end
					io.put_new_line
				end
				cluster_html.put(html_parser, cluster_name(cluster).twin)
			else
				debug
					io.put_string(once " (not readable!)%N")
				end
			end
		end

	add_class (ct: CLASS_TEXT) is
		local
			list: FAST_ARRAY[STRING]; sorter: COLLECTION_SORTER[STRING]; comment: COMMENT; cluster: STRING
		do
			cluster := cluster_name(ct.cluster)
			list := classes.reference_at(cluster)
			if list = Void then
				create list.make(0)
				classes.put(list, cluster.twin)
			end
			sorter.add(list, ct.name.to_string)
			sorter.add(all_classes, ct.name.to_string)
			comment := ct.heading_comment2
			if comment /= Void then
				comments.put(comment, ct.name.to_string)
			end
			core.add(ct)
		end

	write_description (hp: HTML_PARSER; html: HTML_OUTPUT_STREAM; stop_at_dot: BOOLEAN) is
		local
			e: HTML_ELEMENT
		do
			e := body(hp)
			if e /= Void then
				e.to_html_stream(html, stop_at_dot)
				debug
					io.put_string(e.to_string)
					io.put_new_line
				end
			end
		end

	body (hp: HTML_PARSER): EIFFELDOC_HTML is
		local
			i: INTEGER; e: HTML_ELEMENT; node: HTML_NODE
		do
			from
				i := hp.roots.lower
			until
				Result /= Void or else i > hp.roots.upper
			loop
				e := hp.roots.item(i)
				if node ?:= e then
					node ::= e
					node := find_body(node)
					Result := body_node
					Result.set(node)
				end
				i := i + 1
			end
		end

	find_body (node: HTML_NODE): HTML_NODE is
		require
			node /= Void
		local
			i: INTEGER; e: HTML_ELEMENT
		do
			debug
				io.put_string(once "--> ")
				io.put_string(node.name)
				io.put_new_line
			end
			if node.name.same_as(once "body") then
				Result := node
			else
				from
					i := 1
				until
					Result /= Void or else i > node.children_count
				loop
					debug
						io.put_integer(i)
						io.put_new_line
					end
					e := node.child(i)
					if Result ?:= e then
						Result ::= e
						Result := find_body(Result)
					end
					i := i + 1
				end
			end
		end

	put_arrow (html: HTML_OUTPUT_STREAM) is
		local
			s: STRING
		do
			s := options.arrow_address
			if s /= Void then
				html.with_attribute(once "border", once "0")
				html.with_attribute(once "alt", once "->")
				html.put_image(s)
			end
		end

	put_separator (html: HTML_OUTPUT_STREAM) is
		local
			s: STRING
		do
			s := options.separator_address
			if s /= Void then
				html.put_break
				html.with_attribute(once "border", once "0")
				html.with_attribute(once "alt", once "->")
				html.put_image(s)
				html.put_break
			end
		end

end -- class EIFFELDOC
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2005: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
