-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class ID_PROVIDER
	--
	-- Unique object in charge of some id providing.
	--

insert
	GLOBALS

creation {ANY}
	make

feature {C_PRETTY_PRINTER}
	max_id: INTEGER_16
			-- Last already assigned `id'. Maximum 15 bits (see POSITION), so
			-- INTEGER_16 is ok (bit#16 is the sign)

feature {POSITION}
	min_id: INTEGER_16
			-- Last `id' already assigned to a filename.
			--|*** I felt it would be nice not to mix live type `id's and
			--| file name `id's. Therefore, live type `id's grow up from
			--| 0, while file name `id's grow down from
			--| maximum_free_id. However, this makes the code more
			--| complex, and requires two distinct dictionaries. FM

feature {COMPILE_TO_C}
	disk_save is
		local
			i, id: INTEGER_16; tfw: TEXT_FILE_WRITE; name: HASHED_STRING
		do
			create tfw.make
			echo.tfw_connect(tfw, system_tools.id_file_path)
			from
				i := 1
			until
				i > id_memory.count
			loop
				name := id_memory.key(i)
				id := id_memory.item(i)
				tfw.put_integer(id)
				tfw.put_character(' ')
				tfw.put_character('%"')
				tfw.put_string(name.to_string)
				tfw.put_character('%"')
				tfw.put_character(' ')
				smart_eiffel.id_extra_information(tfw, name)
				tfw.put_character('#')
				tfw.put_character('%N')
				i := i + 1
			end
			tfw.disconnect
		end

feature {PARSER, CLASS_TEXT, LIVE_TYPE, C_PRETTY_PRINTER}
	item (hashed_string: HASHED_STRING): INTEGER_16 is
		require
			hashed_string /= Void
		local
			p: POSITION
		do
			if id_memory.fast_has(hashed_string) then
				Result := id_memory.fast_at(hashed_string)
			elseif max_id = min_id - 1 then
				error_handler.append(once "Too many live types (the maximum is ")
				error_handler.append_integer(p.maximum_free_id)
				error_handler.append(once "). Cannot go on: please send a mail at smarteiffel@loria.fr")
				error_handler.print_as_fatal_error
			else
				max_id := max_id + 1
				Result := max_id
				id_memory.add(Result, hashed_string)
			end
			debug
				echo.put_string(once "id(")
				echo.put_string(hashed_string.to_string)
				echo.put_string(once ") = ")
				echo.put_integer(Result)
				echo.put_new_line
			end
		end

feature {EIFFEL_PARSER}
	high_item (hashed_string: HASHED_STRING): INTEGER_16 is
			-- Works like `item', except that if a new ID is allocated,
			-- it is chosen high.
		require
			hashed_string /= Void
		local
			p: POSITION
		do
			if id_memory.fast_has(hashed_string) then
				Result := id_memory.fast_at(hashed_string)
			elseif min_id = max_id + 1 then
				error_handler.append(once "Too many live types (the maximum is ")
				error_handler.append_integer(p.maximum_free_id)
				error_handler.append(once "). Cannot go on: please send a mail at smarteiffel@loria.fr")
				error_handler.print_as_fatal_error
			else
				min_id := min_id - 1
				Result := min_id
				id_memory.add(Result, hashed_string)
			end
			debug
				echo.put_string(once "id(")
				echo.put_string(hashed_string.to_string)
				echo.put_string(once ") = ")
				echo.put_integer(Result)
				echo.put_new_line
			end
		end

feature {POSITION}
	alias_of (id: INTEGER_16): HASHED_STRING is
		do
			Result := id_memory.fast_key_at(id)
			debug
				count_alias_of := count_alias_of + 1
				echo.put_string(once "alias_of(")
				echo.put_integer(id)
				echo.put_string(once ") = ")
				if Result = Void then
					echo.put_string(once "Void")
				else
					echo.put_character('"')
					echo.put_string(Result.to_string)
					echo.put_character('"')
				end
				echo.put_string(once " (")
				echo.put_integer(count_alias_of)
				echo.put_string(once ")%N")
			end
		end

	count_alias_of: INTEGER

feature {}
	id_memory: HASHED_DICTIONARY[INTEGER_16, HASHED_STRING] is
			-- Note: it is not possible to use here a BIJECTIVE_DICTIONARY because, for example INTEGER and INTEGER_32 
			-- share the same id.
		once
			create Result.with_capacity(2048)
		end

	id_memory_add (id: INTEGER_16; string: STRING) is
		require
			string_aliaser.registered_one(string)
		do
			id_memory.add(id, string_aliaser.hashed_string(string))
		end

	make is
		local
			p: POSITION
		do
			id_memory_add(0, as_none)
			id_memory_add(1, as_integer_8)
			id_memory_add(2, as_integer)
			id_memory_add(2, as_integer_32)
			id_memory_add(3, as_character)
			id_memory_add(4, as_real_32)
			id_memory_add(5, as_real)
			id_memory_add(5, as_real_64)
			id_memory_add(6, as_boolean)
			id_memory_add(7, as_string)
			id_memory_add(8, as_pointer)
			id_memory_add(9, as_native_array_character)
			id_memory_add(10, as_integer_16)
			id_memory_add(11, as_integer_64)
			id_memory_add(12, as_real_extended)
			max_id := 12
			min_id := p.maximum_free_id + 1
			if not smart_eiffel.no_id then
				disk_restore
			end
		end

	disk_restore is
		local
			cc: CHARACTER; type_name: STRING; id, item_count: INTEGER_16; tfr: TEXT_FILE_READ; state: INTEGER_8
			predefined_ids: INTEGER; hashed_type_name: HASHED_STRING
		do
			predefined_ids := max_id
			create tfr.make
			echo.tfr_connect(tfr, system_tools.id_file_path)
			if tfr.is_connected then
				from
					if tfr.end_of_input then
						state := 6
					end
				until
					state > 4
				loop
					tfr.read_character
					if tfr.end_of_input then
						state := 5
					else
						cc := tfr.last_character
					end
					inspect
						state
					when 0 then
						-- Waiting first digit of `id'.
						inspect
							cc
						when ' ', '%R', '%N', '%T' then
						when '0' .. '9' then
							id := cc.decimal_value
							state := 1
						else
							state := 6
						end
					when 1 then
						-- Inside `id'.
						inspect
							cc
						when '0' .. '9' then
							id := id * 10 + cc.decimal_value
						when '%"' then
							type_name := temporary_type_name
							type_name.clear_count
							state := 3
						when ' ', '%T' then
							state := 2
						else
							state := 6
						end
					when 2 then
						-- Waiting opening ".
						inspect
							cc
						when '%"' then
							type_name := temporary_type_name
							type_name.clear_count
							state := 3
						when ' ', '%T', '%N', '%R' then
						else
							state := 6
						end
					when 3 then
						-- Inside `type_name'.
						inspect
							cc
						when '%"' then
							item_count := item_count + 1
							if id > predefined_ids then
								hashed_type_name := string_aliaser.hashed_string(type_name)
								id_memory.fast_put(id, hashed_type_name) -- No `add' here because it comes from outside.
								max_id := max_id.max(id)
							end
							state := 4
						when '%N', '%R', '%T' then
							state := 6
						else
							type_name.extend(cc)
						end
					when 4 then
						-- Waiting end of type ('#').
						inspect
							cc
						when '#' then
							state := 0
						else
						end
					else
					end
				end
				tfr.disconnect
				if state = 6 then
					-- Error final state.
					echo.put_string("Corrupted *.id file (after ")
					echo.put_integer(item_count)
					echo.put_string(" correct items).%N")
				else
					check
						state = 5
					end
					-- Final success state.
				end
				echo.put_string(once "Previous IDs reloaded (max_id = ")
				echo.put_integer(max_id)
				echo.put_string(once ").%N")
			end
		end

	temporary_type_name: STRING is
		once
			create Result.make(128)
		end

end -- class ID_PROVIDER
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
