-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
deferred class RUN_FEATURE
	--
	-- A feature at run time: a specialized version of a feature for a given type.
	--
	-- Created at the "adapt" phase, i.e. the latest phase before the back-end generation. An important
	-- property is that during that phase, each run feature is created only once.
	--
	-- * RUN_FEATURE_1: constant attribute.
	-- * RUN_FEATURE_2: attribute.
	-- * RUN_FEATURE_3: procedure.
	-- * RUN_FEATURE_4: function.
	-- * RUN_FEATURE_5: once procedure.
	-- * RUN_FEATURE_6: once function.
	-- * RUN_FEATURE_7: external procedure.
	-- * RUN_FEATURE_8: external function.
	-- * RUN_FEATURE_9: deferred routine.
	--
	--|***PH(20/04/04) should generate assertions for attributes

inherit
	HASHABLE
insert
	GLOBALS

feature {ANY}
	current_type: TYPE_MARK
			-- The type of Current in this feature.
			--|*** should disappear?

	type_of_current: TYPE
			-- The real type of Current in this feature.

	live_type: LIVE_TYPE
			-- Alias the one of `current_type'.
			--|***PH(20/04/04) use type_of_current.live_type instead.

	name: FEATURE_NAME
			-- Final name (the only one really used) of the feature.

	base_feature: ANONYMOUS_FEATURE
			-- Original base feature definition.

	extra_local_expanded_list: FAST_ARRAY[TYPE]
			-- Tmp variables needed for this feature (Void if no one is needed).

	feature_stamp: FEATURE_STAMP is
		do
			Result := type_of_current.lookup(name)
		ensure
			Result.anonymous_feature(type_of_current) = base_feature
		end

	arguments: FORMAL_ARG_LIST is
			-- Runnable arguments list if any.
		deferred
		end

	result_type: TYPE_MARK is
			-- Runnable Result type if any.
		deferred
		end

	require_assertion: E_REQUIRE is
			-- Runnable collected require assertion if any.
		deferred
		end

	local_vars: LOCAL_VAR_LIST is
			-- Runnable local var list if any.
		deferred
		end

	routine_body: INSTRUCTION is
			-- Runnable routine body if any.
		deferred
		end

	ensure_assertion: E_ENSURE is
			-- Runnable collected ensure assertion if any.
		deferred
		end

	rescue_compound: INSTRUCTION is
			-- Runnable rescue compound if any.
		deferred
		end

	is_once_procedure: BOOLEAN is
			-- This is not only True for RUN_FEATURE_5, but it may be also
			-- True when some once procedure is wrapped (RUN_FEATURE_10).
		deferred
		ensure
			Result implies result_type = Void
		end

	is_once_function: BOOLEAN is
			-- This is not only True for RUN_FEATURE_6, but it may be also
			-- True when some once function is wrapped (RUN_FEATURE_11).
		deferred
		ensure
			Result implies result_type /= Void
		end

	is_deferred: BOOLEAN is
		deferred
		end

	hash_code: INTEGER is
		do
			Result := to_pointer.hash_code
		end

	is_root: BOOLEAN
			-- True if `Current' is the root creation feature. Can only be set
			-- once in RUN_FEATURE_3. This function is called via a require
			-- assertion. (This way, no extra memory is used for a production
			-- compiler because this is dead code.)

	frozen is_once_routine: BOOLEAN is
			-- Is it a once routine?
		do
			Result := is_once_function or else is_once_procedure
		end

	side_effect_free: BOOLEAN is
			-- If calling `Current' has no side effect at all.
		require
			smart_eiffel.is_ready
		deferred
		end

	frozen use_current: BOOLEAN is
		require
			smart_eiffel.is_ready
		do
			inspect
				use_current_state
			when True_state then
				Result := True
			when False_state then
			when Unknown_state then
				use_current_state := Computing_state
				if smart_eiffel.scoop then
					if current_type.is_separate then
						use_current_state := True_state
					else
						compute_use_current
					end
				else
					compute_use_current
				end
				Result := use_current_state = True_state
			when Computing_state then
				use_current_state := True_state
				Result := True
			end
		end

	frozen debug_info_in (buffer: STRING) is
			-- For debugging only.
		local
			i: INTEGER
		do
			buffer.extend('{')
			current_type.debug_info_in(buffer)
			buffer.extend('}')
			buffer.extend('.')
			name.complete_name_in(buffer)
			if arguments /= Void then
				buffer.extend('(')
				from
					i := 1
				until
					i > arguments.count
				loop
					arguments.type_mark(i).debug_info_in(buffer)
					i := i + 1
					if i <= arguments.count then
						buffer.extend(',')
					end
				end
				buffer.extend(')')
			end
			if result_type /= Void then
				buffer.extend(':')
				result_type.debug_info_in(buffer)
			end
		end

	frozen start_position: POSITION is
		do
			Result := base_feature.start_position
		end

	c_define_or_scoop_define is
		local
			hlc: BOOLEAN; buffer: STRING
		do
			debug
				buffer := "%T%T"
				name.complete_name_in(buffer)
				buffer.extend('%N')
				echo.put_string(buffer)
			end
			if smart_eiffel.scoop and then current_type.is_separate then
				in_scoop_wrapper := True
				if arguments /= Void then
					hlc := arguments.has_like_current
				end
				define_prototype
				c_scoop_define_wrapper(hlc)
				if result_type = Void then
					cpp.put_string(fz_12)
				else
					cpp.put_string(fz_15)
				end
				-- UNWRAP
				define_unwrap_prototype
				c_scoop_define_unwrap(hlc)
				cpp.put_string(fz_12)
				in_scoop_wrapper := False
			else
				c_define
			end
		end

	c_define is
			-- Produce C code for definition.
		require
			cpp.on_c
		deferred
		ensure
			cpp.on_c
		end

	mapping_c is
			-- Produce C code when current is called and when the concrete type of target is unique (`cpp' is in
			-- charge of the context).
		require
			cpp.on_c
		deferred
		ensure
			cpp.on_c
		end

	mapping_jvm is
		deferred
		end

	frozen id: INTEGER is
		do
			Result := current_type.id
		end

	mapping_name (local_current: BOOLEAN) is
		do
			c_code.clear_count
			if smart_eiffel.scoop and then not local_current and then (current_type.is_separate or else is_root) then
				c_code.append(once "Sw")
			else
				c_code.extend('r')
			end
			id.append_in(c_code)
			name.mapping_c_in(c_code)
			cpp.put_string(c_code)
		end

	frozen jvm_max_locals: INTEGER is
		do
			Result := current_type.jvm_stack_space
			if arguments /= Void then
				Result := Result + arguments.jvm_stack_space(type_of_current)
			end
			if local_vars /= Void then
				Result := Result + local_vars.jvm_stack_space(type_of_current)
			end
			if result_type /= Void then
				Result := Result + result_type.jvm_stack_space
			end
			if rescue_compound /= Void then
				Result := Result + 1
			end
		end

feature {ADDRESS_OF_POOL}
	address_of_c_define (caller: ADDRESS_OF) is
			-- Corresponding `caller' is used for error messages.
		require
			caller /= Void
		deferred
		end

feature {ADDRESS_OF}
	address_of_c_mapping is
			-- Produce C code for operator $<feature_name>
		require
			live_type.at_run_time
			cpp.on_c
		deferred
		ensure
			cpp.on_c
		end

feature {NATIVE}
	frozen default_mapping_procedure is
			-- Default mapping for procedure calls with target.
		do
			default_mapping_function
			cpp.put_string(fz_00)
		end

	frozen default_mapping_function is
			-- Default mapping for function calls with target.
		local
			no_check, uc, tcbd: BOOLEAN
		do
			no_check := ace.no_check
			uc := use_current
			if not uc then
				tcbd := cpp.target_cannot_be_dropped
				if tcbd then
					cpp.put_character(',')
				end
			end
			mapping_name(True)
			default_mapping_arg(no_check, uc, tcbd)
		end

	frozen default_mapping_arg (no_check, uc, tcbd: BOOLEAN) is
		do
			cpp.put_character('(')
			if no_check then
				cpp.put_string(once "&ds")
			end
			if ace.profile then
				if no_check then
					cpp.put_character(',')
				end
				cpp.put_string(once "&local_profile")
			end
			if uc then
				if no_check or else ace.profile then
					cpp.put_character(',')
				end
				if current_type.is_boolean then
					cpp.put_string(once "(T6)(")
				end
				cpp.put_target_as_target
				if current_type.is_boolean then
					cpp.put_character(')')
				end
			end
			if arguments /= Void then
				if uc or else no_check or else ace.profile then
					cpp.put_character(',')
				end
				cpp.put_arguments
			end
			cpp.put_character(')')
			if not uc and then tcbd then
				cpp.put_character(')')
			end
		end

	routine_mapping_jvm is
		local
			rt, ct: TYPE_MARK; idx, stack_level: INTEGER
		do
			ct := current_type
			jvm.push_target_as_target
			stack_level := - (1 + jvm.push_arguments)
			rt := result_type
			if rt /= Void then
				stack_level := stack_level + rt.jvm_stack_space
			end
			idx := constant_pool.idx_methodref(Current)
			ct.live_type.jvm_invoke(idx, stack_level)
		end

	frozen c_define_with_body (body: STRING) is
		require
			body /= Void
		do
			define_prototype
			c_define_opening
			cpp.put_string(body)
			c_define_closing
			if result_type = Void then
				cpp.put_string(fz_12)
			else
				cpp.put_string(fz_15)
			end
		end

feature {LIVE_TYPE}
	jvm_field_or_method is
			-- Update jvm's `fields' or `methods' if needed.
		deferred
		end

	scoop_expanded_verified: BOOLEAN
			-- Anti-recursion flag used by `verify_scoop_expanded'

	verify_scoop_expanded (p: POSITION) is
			-- default does nothing. Redefined in the query-type run_features (those having a non-Void result
			-- type) to call `default_verify_scoop_expanded'
		do
			scoop_expanded_verified := True
		ensure
			scoop_expanded_verified
		end

feature {}
	default_verify_scoop_expanded (p: POSITION) is
		require
			result_type /= Void
		local
			t: TYPE_MARK
		do
			t := result_type
			if not t.is_expanded then
				error_handler.add_position(start_position)
				error_handler.add_position(p)
				fatal_error(fz_sc4)
			elseif not t.live_type.scoop_expanded_verified then
				t.live_type.verify_scoop_expanded(p)
			end
			scoop_expanded_verified := True
		ensure
			scoop_expanded_verified
		end

feature {CONSTANT_POOL, SWITCH_COLLECTION}
	frozen jvm_descriptor: STRING is
		do
			tmp_jvm_descriptor.clear_count
			update_tmp_jvm_descriptor
			Result := tmp_jvm_descriptor
		end

feature {SWITCH, C_PRETTY_PRINTER, GC_HANDLER, CECIL_POOL}
	c_set_dump_stack_top_in (code_buffer, ds, comment: STRING) is
		do
			cpp.c_set_dump_stack_top_for_in(current_type.type, code_buffer, ds, comment)
		end

	c_set_dump_stack_top (ds, comment: STRING) is
		require
			cpp.on_c
		do
			c_code.clear_count
			c_set_dump_stack_top_in(c_code, ds, comment)
			cpp.put_string(c_code)
		ensure
			cpp.on_c
		end

feature {LIVE_TYPE}
	prepare_introspection (put_else: BOOLEAN): BOOLEAN is
		require
			cpp.on_c
			ace.no_check
		do
			Result := put_else
		ensure
			put_else implies Result
		end

	prepare_introspection2 (put_coma: BOOLEAN): BOOLEAN is
		require
			cpp.on_c
			ace.no_check
		do
			Result := put_coma
		ensure
			put_coma implies Result
		end

feature {JVM}
	jvm_define is
			-- To compute the constant pool, the number of fields, the number of methods, etc.
		require
			smart_eiffel.is_ready
		deferred
		end

	frozen jvm_argument_offset (a: ARGUMENT_NAME1): INTEGER is
		require
			arguments /= Void
		do
			Result := current_type.jvm_stack_space
			Result := Result + arguments.jvm_offset_of(type_of_current, a)
		ensure
			Result >= a.rank - 1
		end

	frozen jvm_local_variable_offset (ln: LOCAL_NAME1): INTEGER is
		require
			local_vars /= Void
		do
			Result := current_type.jvm_stack_space
			if arguments /= Void then
				Result := Result + arguments.jvm_stack_space(type_of_current)
			end
			Result := Result + local_vars.jvm_offset_of(type_of_current, ln)
		ensure
			Result >= ln.rank - 1
		end

feature {RUN_FEATURE}
	is_in_computation: BOOLEAN is
		do
			Result := use_current_state = Computing_state
		end

feature {RUN_FEATURE, E_RETRY}
	jvm_try_begin: INTEGER

	jvm_try_end: INTEGER

	jvm_handler: INTEGER

feature {}
	frozen put_c_name_tag is
		require
			ace.no_check
		local
			fn: FEATURE_NAME; tag: STRING
		do
			--|*** Afficher tous les synonymes avec une boucle ??? ***.
			fn := base_feature.first_name
			if not fn.is_simple_feature_name then
				tag := name.infix_or_prefix
				if tag /= Void then
					cpp.put_character('%"')
					cpp.put_string(tag)
					cpp.put_character(' ')
					cpp.put_character('%"')
				end
				cpp.put_string_c(name.to_string)
				cpp.put_string(once "%" (%"")
			end
			tag := fn.infix_or_prefix
			if tag /= Void then
				cpp.put_character('%"')
				cpp.put_string(tag)
				cpp.put_character(' ')
				cpp.put_character('%"')
			end
			cpp.put_string_c(fn.to_string)
			cpp.put_character('%"')
			cpp.put_character(' ')
			cpp.put_string(current_type.written_mark)
			if not fn.is_simple_feature_name then
				cpp.put_character(')')
			end
			cpp.put_character('%"')
		end

	address_of_wrapper_name_in (str: STRING) is
		do
			str.extend('W')
			id.append_in(str)
			name.mapping_c_in(str)
		end

	address_of_c_define_wrapper (caller: ADDRESS_OF) is
		require
			cpp.on_c
		do
			c_code.clear_count
			if result_type = Void then
				c_code.append(fz_void)
			else
				result_type.c_type_for_external_in(c_code)
			end
			c_code.extend(' ')
			address_of_wrapper_name_in(c_code)
			c_code.extend('(')
			current_type.c_type_for_external_in(c_code)
			c_code.extend(' ')
			c_code.extend('C')
			if arguments /= Void then
				c_code.extend(',')
				arguments.external_prototype_in(c_code)
			end
			c_code.extend(')')
			cpp.put_c_heading(c_code)
			cecil_pool.define_body_of(Current)
		end

	address_of_c_mapping_wrapper is
		do
			c_code.clear_count
			address_of_wrapper_name_in(c_code)
			cpp.put_string(c_code)
		end

feature {}
	c_frame_descriptor is
			-- Initialize all `c_frame_descriptor' variables accordingly.
		require
			ace.no_check
		local
			c_frame_descriptor_format, c_frame_descriptor_locals, oresult: STRING; t: TYPE_MARK
		do
			c_frame_descriptor_format := cpp.c_frame_descriptor_format
			c_frame_descriptor_locals := cpp.c_frame_descriptor_locals
			c_frame_descriptor_format.clear_count
			c_frame_descriptor_locals.clear_count
			if use_current then
				current_type.c_frame_descriptor_in(c_frame_descriptor_format)
			end
			if arguments /= Void then
				arguments.c_frame_descriptor(type_of_current, c_frame_descriptor_format, c_frame_descriptor_locals)
			end
			if is_once_function then
				oresult := once_routine_pool.o_result(base_feature)
				t := result_type
				c_frame_descriptor_locals.append(once "(void**)&")
				c_frame_descriptor_locals.append(oresult)
				c_frame_descriptor_locals.extend(',')
				c_frame_descriptor_format.append(as_result)
				t.c_frame_descriptor_in(c_frame_descriptor_format)
			elseif result_type /= Void then
				t := result_type
				c_frame_descriptor_locals.append(once "(void**)&R,")
				c_frame_descriptor_format.append(as_result)
				t.c_frame_descriptor_in(c_frame_descriptor_format)
			end
			if local_vars /= Void then
				local_vars.c_frame_descriptor(type_of_current, c_frame_descriptor_format, c_frame_descriptor_locals)
			end
		end

	define_prototype is
		require
			live_type.at_run_time
			cpp.on_c
		local
			mem_id: INTEGER; no_check, profile: BOOLEAN
		do
			no_check := ace.no_check
			profile := ace.profile
			mem_id := id -- Define heading of corresponding C function.
			c_code.clear_count
			-- Extra comment to debug C code :
			c_code.append(once "/*")
			c_code.append(current_type.written_mark)
			c_code.append(once "*/%N")
			--
			if result_type = Void then
				c_code.append(fz_void)
			else
				result_type.c_type_for_result_in(c_code)
			end
			c_code.extend(' ')
			c_code.extend('r')
			mem_id.append_in(c_code)
			name.mapping_c_in(c_code)
			c_code.extend('(')
			if no_check then
				c_code.append(once "se_dump_stack*caller")
				if profile or else use_current or else arguments /= Void then
					c_code.extend(',')
				end
			end
			if profile then
				c_code.append(once "se_local_profile_t*parent_profile")
				if use_current or else arguments /= Void then
					c_code.extend(',')
				end
			end
			if use_current then
				current_type.c_type_for_target_in(c_code)
				c_code.append(once " C")
				if arguments /= Void then
					c_code.extend(',')
				end
			end
			if arguments = Void then
				if no_check or else profile then
				elseif not use_current then
					c_code.append(fz_void)
				end
			else
				arguments.compile_to_c_in(type_of_current, c_code)
			end
			c_code.extend(')')
			cpp.put_c_heading(c_code)
			cpp.swap_on_c
		ensure
			cpp.on_c
		end

	define_unwrap_prototype is
		require
			live_type.at_run_time
			cpp.on_c
		local
			mem_id: INTEGER; no_check: BOOLEAN
		do
			no_check := ace.no_check
			mem_id := id -- Define heading of corresponding C function.
			c_code.clear_count
			-- Extra comment to debug C code :
			c_code.append(once "/*")
			c_code.append(current_type.written_mark)
			c_code.append(once "*/%N")
			--
			c_code.append(once "void Su")
			mem_id.append_in(c_code)
			name.mapping_c_in(c_code)
			c_code.append(once "(void* data, int data_length, void* result_ref)")
			cpp.put_c_heading(c_code)
			cpp.swap_on_c
		ensure
			cpp.on_c
		end

	old_list: FAST_ARRAY[E_OLD]
			-- Non Void when `ensure_assertion' actually has "old" expression.

	c_define_opening is
			-- Define opening section in C function.
		require
			--not current_type.is_separate or else current_type.class_text.is_separate
		local
			t: TYPE_MARK; no_check: BOOLEAN; oresult: STRING; rf: RUN_FEATURE;
			clc, i, result_type_id, class_invariant_flag: INTEGER
		do
			smart_eiffel.push_context(base_feature)
			no_check := ace.no_check
			if no_check then
				c_frame_descriptor
			end
			-- (0) --------------------------- Exception handling :
			if rescue_compound /= Void then
				if smart_eiffel.scoop then
					error_handler.add_position(rescue_compound.start_position)
					fatal_error("Exceptions cannot currently be used with SCOOP.")
				end
				cpp.put_string(once "struct rescue_context rc;%N")
			end
			-- (0') ---------------------------- Current subsystem:
			if current_type.is_separate then
				cpp.put_string(once "se_subsystem_t* self=se_current_subsystem_thread();%N")
			end
			-- (1) -------------------- Local variable for Result:
			if is_once_function then
				oresult := once_routine_pool.o_result(base_feature)
			elseif result_type /= Void then
				t := result_type
				c_code.clear_count
				t.c_type_for_result_in(c_code)
				c_code.append(once " R=")
				t.c_initialize_in(c_code)
				c_code.append(fz_00)
				cpp.put_string(c_code)
			end
			-- (2) ----------------------- User's local variables:
			if local_vars /= Void then
				local_vars.c_declare(type_of_current, rescue_compound /= Void)
			end
			-- (3) ------------------- Local variables for profile:
			if ace.profile then
				smart_eiffel.local_profile
			end
			-- (4) ----------- Local variables for old expressions:
			if old_list /= Void then
				check
					ensure_assertion /= Void
				end
				from
					i := old_list.upper
				until
					i < old_list.lower
				loop
					old_list.item(i).c_declare_for_old(type_of_current)
					i := i - 1
				end
			end
			-- (5) ------------- Temporary variables for expanded:
			cpp.put_extra_local_expanded(extra_local_expanded_list)
			if no_check then
				-- (6) ------------------------------- Prepare locals:
				clc := cpp.c_frame_descriptor_format.occurrences('%%')
				if clc > 0 then
					clc := clc #// 2
					if use_current then
						clc := clc - 1
					end
				end
				if clc > 0 then
					cpp.put_string(once "void**locals[")
					cpp.put_integer(clc)
					cpp.put_string(once "];%N")
				end
				-- (7) ----------------------------------- Prepare ds:
				c_initialize_frame_information(clc)
				-- (8) ------------------------ Initialise Dump Stack:
				c_set_dump_stack_top(once "&ds", fz_link)
			end
			-- (9) ---------------------------------- Add profile:
			if ace.profile then
				smart_eiffel.start_profile(Current)
			end
			-- (10) --------------------- Execute old expressions:
			if old_list /= Void then
				from
					i := old_list.upper
				until
					i < old_list.lower
				loop
					old_list.item(i).compile_to_c_old(type_of_current)
					i := i - 1
				end
			end
			-- (11) -------------------------- Exception handling :
			if rescue_compound /= Void then
				cpp.put_string(once "handle(SE_HANDLE_EXCEPTION_SET,NULL);if(SETJMP(rc.jb)!=0){/*rescue*/%N")
				rescue_compound.compile_to_c(type_of_current)
				cpp.put_string(once "internal_exception_handler(Routine_failure);%N}%N")
			end
			-- (12) ------------ Initialize Result/local expanded :
			if result_type /= Void then
				if result_type.is_user_expanded then
					rf := result_type.live_type.default_create_run_feature
					if rf /= Void then
						result_type_id := result_type.live_type.id
						cpp.put_string(once "{T")
						cpp.put_integer(result_type_id)
						cpp.put_string(once " n=M")
						cpp.put_integer(result_type_id)
						cpp.put_string(once ";%N")
						cpp.push_create_instruction(type_of_current, rf, Void)
						rf.mapping_c
						cpp.pop
						if oresult /= Void then
							cpp.put_string(oresult)
						else
							cpp.put_string(once "R")
						end
						cpp.put_string(once "=n;}%N")
					end
					-- Even when there is no default creation procedure to apply, we must call the class 
					-- invariant:
					class_invariant_flag := cpp.class_invariant_call_opening(result_type.type, True)
					if class_invariant_flag > 0 then
						if result_type.need_c_struct then
							cpp.put_character('&')
						end
						if oresult /= Void then
							cpp.put_string(oresult)
						else
							cpp.put_character('R')
						end
						cpp.class_invariant_call_closing(class_invariant_flag, True)
					end
				end
			end
			if local_vars /= Void then
				local_vars.initialize_expanded(type_of_current)
			end
			-- (13) --------------------------- Retry start label :
			if rescue_compound /= Void then
				cpp.put_string(once "retry_tag:%N")
			end
			-- (14) ---------------------- Require assertion code :
			if current_type.class_text.require_check and then require_assertion /= Void then
				require_assertion.compile_to_c(type_of_current)
			end
			-- (15) ------------------------- Save rescue context :
			if rescue_compound /= Void then
				cpp.put_string(once "rc.next = rescue_context_top;%N%
                           %rescue_context_top = &rc;%N")
				if no_check then
					cpp.put_string(once "rc.top_of_ds=&ds;%N")
					c_set_dump_stack_top(once "&ds", fz_link)
				end
			end
			-- (16) ------------------------------ SCOOP handling :
			if smart_eiffel.scoop and then arguments /= Void then
				arguments.c_lock_scoop_init(type_of_current)
				if require_assertion = Void then
					arguments.c_lock_scoop
				else
					require_assertion.c_scoop_while
					arguments.c_lock_scoop
					require_assertion.c_scoop_if_not_guard
					arguments.c_unlock_scoop
					require_assertion.c_scoop_guard
				end
				cpp.put_string(once "{%N")
			end
		end

	c_define_closing is
			-- Define closing section in the corresponding C function (code for ensure checking, free memory of
			-- expanded, run stack pop, etc.).
		local
			sedb_trace_before_exit: POSITION
		do
			-- (1) --------------------------- Ensure Check Code:
			if ensure_assertion /= Void then
				ensure_assertion.compile_to_c(type_of_current)
			end
			-- (2) ----------------------------- Class Invariant:
			if use_current then
				if name.to_string /= as_dispose then
					cpp.current_class_invariant(current_type)
				end
			end
			-- (3) ---------------------------------- For rescue:
			if rescue_compound /= Void then
				cpp.put_string(once "rescue_context_top = rc.next;handle(SE_HANDLE_EXCEPTION_CLEAR,NULL);%N")
			end
			-- (4) ------------------------------- Run Stack Pop:
			if ace.no_check then
				sedb_trace_before_exit := base_feature.sedb_trace_before_exit
				if not sedb_trace_before_exit.is_unknown then
					cpp.put_trace_or_sedb_instruction(sedb_trace_before_exit)
				end
				c_set_dump_stack_top(fz_caller, fz_unlink)
			end
			-- (5) ------------------------------------ Profiling:
			if ace.profile then
				smart_eiffel.stop_profile
			end
			-- (6) ------------------------------- SCOOP handling:
			if smart_eiffel.scoop and then arguments /= Void and then arguments.c_lock_scoop_count > 0 then
				cpp.put_string(once "}%N")
				arguments.c_unlock_scoop
				cpp.put_string(once "/*</SCOOP>*/}%N")
			end
			smart_eiffel.pop_context(base_feature)
		end

	use_current_state: INTEGER
			-- Memory cache for `use_current' (see below *_state constants).

	False_state,    -- already computed and False.
	True_state,     -- already computed and True.
	Unknown_state,  -- not yet computed.
	Computing_state -- during computation.
	: INTEGER is unique

	frozen std_compute_use_current is
		require
			is_in_computation
		do
			smart_eiffel.push_context(base_feature)
			if use_current_state = Computing_state then
				if require_assertion /= Void then
					if require_assertion.use_current(type_of_current) then
						use_current_state := True_state
					end
				end
			end
			if use_current_state = Computing_state then
				if routine_body /= Void then
					if routine_body.use_current(type_of_current) then
						use_current_state := True_state
					end
				end
			end
			if use_current_state = Computing_state then
				if rescue_compound /= Void then
					if rescue_compound.use_current(type_of_current) then
						use_current_state := True_state
					end
				end
			end
			if use_current_state = Computing_state then
				if ensure_assertion /= Void then
					if ensure_assertion.use_current(type_of_current) then
						use_current_state := True_state
					end
				end
			end
			if use_current_state = Computing_state then
				use_current_state := False_state
			end
			smart_eiffel.pop_context(base_feature)
		ensure
			use_current_state = False_state or else use_current_state = True_state
		end

	compute_use_current is
		require
			is_in_computation
		deferred
		ensure
			use_current_state = True_state or else use_current_state = False_state
		end

	c_code: STRING is
		once
			create Result.make(512)
		end

	update_tmp_jvm_descriptor is
		deferred
		end

	tmp_jvm_descriptor: STRING is
		once
			create Result.make(128)
		end

	routine_update_tmp_jvm_descriptor is
			-- For RUN_FEATURE_3/4/5/6/7/8/9/10/11 :
		local
			ct, rt: TYPE_MARK
		do
			tmp_jvm_descriptor.extend('(')
			ct := current_type
			ct.jvm_target_descriptor_in(tmp_jvm_descriptor)
			if arguments /= Void then
				arguments.jvm_descriptor_in(tmp_jvm_descriptor)
			end
			rt := result_type
			if rt = Void then
				tmp_jvm_descriptor.append(fz_19)
			else
				tmp_jvm_descriptor.extend(')')
				rt.jvm_descriptor_in(tmp_jvm_descriptor)
			end
		end

	method_info_start is
		local
			flags: INTEGER; buffer: STRING

		do
			flags := current_type.jvm_method_flags
			buffer := once "..............."
			buffer.clear_count
			name.mapping_c_in(buffer)
			method_info.start(flags, buffer, jvm_descriptor)
		end

	jvm_increment_invariant_flag is
		local
			name_idx: INTEGER; descriptor_idx: INTEGER; idx: INTEGER
		do
			if not method_info.is_static and live_type.class_text.invariant_check and not live_type.invariant_flag then
				name_idx := constant_pool.idx_utf8(once "invariant_flag")
				descriptor_idx := constant_pool.idx_utf8(fz_i)
				field_info.add(1, name_idx, descriptor_idx)
				live_type.set_invariant_flag
			end
			if not method_info.is_static and live_type.class_text.invariant_check then
				code_attribute.opcode_aload_0
				live_type.opcode_checkcast_1
				idx := constant_pool.idx_fieldref3(live_type.fully_qualified_name, once "invariant_flag", fz_i)
				code_attribute.opcode_getfield(idx, -1)
				code_attribute.opcode_iconst_1
				code_attribute.opcode_iadd
				code_attribute.opcode_aload_0
				live_type.opcode_checkcast_1
				code_attribute.opcode_swap
				code_attribute.opcode_putfield(idx, 1)
			end
		end

	jvm_decrement_invariant_flag is
		local
			idx: INTEGER
		do
			if not method_info.is_static and live_type.class_text.invariant_check then
				code_attribute.opcode_aload_0
				live_type.opcode_checkcast_1
				idx := constant_pool.idx_fieldref3(live_type.fully_qualified_name, once "invariant_flag", fz_i)
				code_attribute.opcode_getfield(idx, -1)
				code_attribute.opcode_iconst_1
				code_attribute.opcode_isub
				code_attribute.opcode_aload_0
				live_type.opcode_checkcast_1
				code_attribute.opcode_swap
				code_attribute.opcode_putfield(idx, 1)
			end
		end

	jvm_define_opening is
		require
			jvm.current_frame = Current
		local
			space, i: INTEGER
		do
			-- (1) -------------------- Local variable for Result :
			if result_type /= Void then
				space := result_type.jvm_push_default
				if is_once_function then
					once_routine_pool.jvm_result_store(type_of_current, base_feature)
				else
					result_type.jvm_write_local_creation(base_feature.jvm_result_offset(type_of_current))
				end
			end
			-- (2) ----------------------- User's local variables :
			if local_vars /= Void then
				local_vars.jvm_initialize(type_of_current)
			end
			-- (3) ------------------------ Execute old epressions:
			if old_list /= Void then
				check
					ensure_assertion /= Void
				end
				from
					i := old_list.upper
				until
					i < old_list.lower
				loop
					old_list.item(i).compile_to_jvm_old(type_of_current)
					i := i - 1
				end
			end
			jvm_try_begin := code_attribute.program_counter -- (4) ----------------------- Require assertion code :
			if current_type.class_text.require_check and then require_assertion /= Void then
				require_assertion.compile_to_jvm(type_of_current)
			end
		end

	jvm_define_closing is
		require
			jvm.current_frame = Current
		local
			idx: INTEGER
		do
			-- (0) ----------------------------- Class Invariant :
			-- (1) --------------------------- Ensure Check Code :
			if live_type.class_text.ensure_check then
				if ensure_assertion /= Void then
					ensure_assertion.compile_to_jvm(type_of_current, True)
					code_attribute.opcode_pop
				end
			end
			jvm_try_end := code_attribute.program_counter
			live_type.jvm_check_class_invariant_1
			-- (2) --------------------- Free for local expanded :
			-- (3) --------------------- Rescue :
			jvm_handler := code_attribute.program_counter + 1
			if rescue_compound /= Void then
				-- save exception
				code_attribute.opcode_return
				code_attribute.opcode_astore(jvm_max_locals - 1)
				-- rescue clause
				rescue_compound.compile_to_jvm(type_of_current)
				-- rethrow exception
				code_attribute.opcode_aload(jvm_max_locals - 1)
				code_attribute.opcode_athrow
				idx := constant_pool.idx_class2(once "java/lang/Exception")
				code_attribute.add_exception(jvm_try_begin, jvm_try_end, jvm_handler, idx)
			end
		end

	c_initialize_frame_information (c_locals_count: INTEGER) is
		require
			ace.no_check
		local
			c_frame_descriptor_format, c_frame_descriptor_locals: STRING; i, j: INTEGER; c: CHARACTER
		do
			c_frame_descriptor_format := cpp.c_frame_descriptor_format
			c_frame_descriptor_locals := cpp.c_frame_descriptor_locals
			cpp.put_string(once "static se_frame_descriptor fd={")
			put_c_name_tag
			c_code.clear_count
			c_code.extend(',')
			if use_current then
				c_code.extend('1')
			else
				c_code.extend('0')
			end
			c_code.extend(',')
			c_locals_count.append_in(c_code)
			c_code.append(once ",%"")
			c_code.append(c_frame_descriptor_format)
			c_code.append(once "%",1};%Nse_dump_stack ds;%Nds.fd=&fd;%N%
			    %ds.current=")
			if use_current then
				if current_type.is_expanded then
					if current_type.is_user_expanded then
						c_code.append(once "(void**)")
						--elseif smart_eiffel.live_type(current_type).run_time_mark = as_native_array_character then
						--   -- NATIVE_ARRAY[CHARACTER]
						--   c_code.append(once "(void*)")
					else
						c_code.append(once "(void**)&")
					end
				else
					c_code.append(once "(void**)&")
				end
				c_code.extend('C')
			else
				c_code.append(fz_null)
			end
			c_code.extend(';')
			c_code.extend('%N')
			cpp.put_string(c_code)
			cpp.put_position_in_ds(start_position)
			cpp.put_string(once "ds.caller=caller;%N")
			if c_locals_count <= 0 then
				cpp.put_string(once "ds.locals=NULL;%N")
			else
				cpp.put_string(once "ds.locals=locals;%N")
			end
			cpp.put_string(once "ds.exception_origin=NULL;%N")
			from
				j := 1
			until
				i = c_locals_count
			loop
				cpp.put_string(once "locals[")
				cpp.put_integer(i)
				cpp.put_string(once "]=")
				from
					c := c_frame_descriptor_locals.item(j)
				until
					c = ','
				loop
					cpp.put_character(c)
					j := j + 1
					c := c_frame_descriptor_locals.item(j)
				end
				j := j + 1
				cpp.put_string(fz_00)
				i := i + 1
			end
		end

	c_scoop_initialize_frame_descriptor (n: STRING) is
		require
			ace.no_check
		do
			cpp.put_string(once "static se_frame_descriptor fd={")
			cpp.put_string(n)
			put_c_name_tag
			cpp.put_string(once "[
	    ,0,0,"",1};
	    se_dump_stack ds = {NULL,NULL,0,NULL,NULL};
	    ds.fd=&fd;

		   ]")
			cpp.put_position_in_ds(start_position)
			--|*** MUST NOT BE NULL FOR SEPARATE DEBUGGING
			c_set_dump_stack_top(once "&ds", fz_link)
		end

	c_scoop_define_wrapper (has_like_current: BOOLEAN) is
			-- Define the SCOOP function wrapper. (The C prototype is the same as
			-- an ordinary function.)
		require
			cpp.on_c
			smart_eiffel.scoop
		local
			i, c, mem_id, local_id: INTEGER; t, local_type: TYPE_MARK; local_run_feature: RUN_FEATURE
		do
			mem_id := id
			if has_like_current then
				local_type := current_type
				local_id := mem_id
				local_run_feature := Current
			else
				local_type := current_type.local_from_separate
				local_id := local_type.id
				local_run_feature := local_type.live_type.at(name)
			end
			check
				local_run_feature /= Void
			end
			-- Struct for the local variables to wrap:
			c_code.copy(once "typedef struct {")
			local_type.c_type_for_target_in(c_code)
			c_code.append(once " C;")
			if arguments /= Void then
				arguments.c_scoop_struct_in(type_of_current, c_code)
			end
			c_code.append(once "} Tw")
			local_id.append_in(c_code)
			name.mapping_c_in(c_code)
			c_code.append(fz_00)
			cpp.put_string_on_h(c_code)
			-- (0) ------------------------ Local SCOOP variables:
			cpp.put_string(once "[
	    se_subsystem_t* self = se_current_subsystem_thread();
	    se_subsystem_t* client = C->subsystem;

			      ]")
			-- (1) -------------------- Local variable for Result :
			if result_type /= Void then
				t := result_type
				c_code.clear_count
				t.c_type_for_result_in(c_code)
				c_code.append(once " R=")
				t.c_initialize_in(c_code)
				c_code.append(fz_00)
				cpp.put_string(c_code)
			end
			-- (2) ----------------------------------- Prepare ds:
			if ace.no_check then
				c_scoop_initialize_frame_descriptor(once "%"<separate wrapper> %"")
			end
			-- (3) --------------------- False separate detection:
			cpp.put_string(once "if (self == client) {%N")
			if has_like_current then
				c_define_body
			else
				if result_type /= Void then
					cpp.put_string(fz_26)
				end
				cpp.push_scoop_wrapper(type_of_current, local_run_feature)
				local_run_feature.mapping_c
				cpp.pop
				if result_type /= Void then
					cpp.put_string(fz_14)
				end
			end
			cpp.put_string(once "}%Nelse {%N")
			c_code.copy(once "Tw")
			local_id.append_in(c_code)
			name.mapping_c_in(c_code)
			c_code.append(once "* data = (Tw")
			local_id.append_in(c_code)
			name.mapping_c_in(c_code)
			c_code.append(once "*)malloc(sizeof(Tw")
			local_id.append_in(c_code)
			name.mapping_c_in(c_code)
			c_code.append(once "));%N")
			cpp.put_string(c_code)
			-- (4) -------------------------------- Function wrap:
			if use_current then
				c_code.clear_count
				if has_like_current then
					c_code.append(once "data->C=C;%N")
				else
					c_code.append(once "data->C=C->ref;%N")
					--|*** TO BE CHECKED for CLASS_TYPE_MARK
				end
				cpp.put_string(c_code)
			end
			if arguments /= Void then
				from
					i := 1
					c := arguments.count
				until
					i > c
				loop
					c_code.copy(once "data->_")
					c_code.append(arguments.name(i).to_string)
					if arguments.type_mark(i).to_static(type_of_current).is_expanded then
						c_code.append(once "=(a")
						i.append_in(c_code)
					else
						c_code.append(once "=(T0*)as_separate(self,a")
						i.append_in(c_code)
					end
					c_code.append(fz_14)
					cpp.put_string(c_code)
					i := i + 1
				end
			end
			-- (5) -------------------------------- Function call:
			c_code.clear_count
			if result_type /= Void then
				c_code.append(once "client->vft.query")
			else
				c_code.append(once "client->vft.command")
			end
			c_code.append(once "(client,&Su")
			mem_id.append_in(c_code)
			name.mapping_c_in(c_code)
			c_code.append(once ",data,sizeof(Tw")
			local_id.append_in(c_code)
			name.mapping_c_in(c_code)
			c_code.extend(')')
			if result_type /= Void then
				c_code.append(once ",&R")
			end
			c_code.append(fz_14)
			cpp.put_string(c_code)
			cpp.put_string(fz_12)
			-- (5) ------------------------------------ Unlink ds:
			if ace.no_check then
				c_set_dump_stack_top(fz_caller, fz_unlink)
			end
		end

	c_scoop_define_unwrap (has_like_current: BOOLEAN) is
			-- Define function unwrapper
			-- C prototype: void unwrap(void* data, int length, void* result)
		require
			cpp.on_c
			smart_eiffel.scoop
		local
			mem_id: INTEGER; local_type: TYPE_MARK; local_run_feature: RUN_FEATURE
		do
			mem_id := id -- (0) ------------------------ Local SCOOP variables:
			cpp.put_string(once "se_subsystem_t* self = se_current_subsystem_thread();%N")
			-- (2) ----------------------------------- Prepare ds:
			if ace.no_check then
				c_scoop_initialize_frame_descriptor(once "%"<separate unwrapper> %"")
			end
			-- (3) -------------------------------- Function call:
			if result_type /= Void then
				c_code.copy(once "*((")
				result_type.c_type_for_result_in(c_code)
				c_code.append(once "*)result_ref)=")
				cpp.put_string(c_code)
			end
			if has_like_current then
				cpp.push_scoop_unwrapper(type_of_current, Current)
				mapping_c
				cpp.pop
			else
				local_type := current_type.local_from_separate
				local_run_feature := local_type.live_type.at(name)
				check
					local_type.live_type.run_time_set.count = 1
				end
				cpp.push_scoop_unwrapper(type_of_current, local_run_feature)
				--|*** Not sure. (Dom. march 30th 2004) ***
				local_run_feature.mapping_c
				cpp.pop
			end
			if result_type /= Void then
				cpp.put_string(fz_00)
				if result_type.is_reference then
					cpp.put_string(once "*((sT0**)result_ref)=as_separate(self,%
				   %*(T0**)result_ref);%N")
				end
			end
			cpp.put_string(once "free(data);%N")
			-- (4) ------------------------------------ Unlink ds:
			if ace.no_check then
				c_set_dump_stack_top(fz_null, fz_unlink)
			end
		end

	frozen for (lt: LIVE_TYPE; origin: like base_feature; fn: FEATURE_NAME) is
			-- Creation feature for the new lookup. Just creates the run feature, without adapting it yet.
		require
			not smart_eiffel.is_ready
			origin /= Void
		local
			buffer: STRING
		do
			live_type := lt
			base_feature := origin
			name := fn
			type_of_current := lt.type
			current_type := type_of_current.canonical_type_mark --|*** 
			use_current_state := Unknown_state
			debug
				buffer := "Creating "
				buffer.append(current_type.written_mark)
				buffer.extend('.')
				name.complete_name_in(buffer)
				buffer.extend('%N')
				echo.put_string(buffer)
			end
			set_result_type
			smart_eiffel.register_run_feature(Current)
		end

	set_result_type is
		deferred
		end

	hook is
			-- Add some hooks to the initialization. Currently only used by RUN_FEATURE_6 to initialize the
			-- special STD_OUTPUT.flush
			--|*** CAD: who can give a better solution?
		do
		end

	debug_info: STRING -- To view more under sedb.

	do_adapt is
			-- Called by `adapt'.
		deferred
		end

feature {LIVE_TYPE}
	frozen adapt is
			-- Adapt the feature to make it suitable for use by the good old C_PRETTY_PRINTER.
		require
			not adapted
			smart_eiffel.registered(Current)
			type_of_current.live_type.registered(Current)
		do
			adapted := True -- do that first because of recursive features
			debug
				debug_info := "<adapting> {"
				current_type.debug_info_in(debug_info)
				debug_info.extend('}')
				debug_info.extend('.')
				name.complete_name_in(debug_info)
			end
			base_feature.hook_for(live_type)
			check
				smart_eiffel.get_and_clear_extra_local_expanded_list = Void
			end
			smart_eiffel.push_context(base_feature)
			do_adapt
			smart_eiffel.pop_context(base_feature)
			extra_local_expanded_list := smart_eiffel.get_and_clear_extra_local_expanded_list
			debug
				debug_info.clear_count
				debug_info_in(debug_info)
			end
			--if current_type.is_separate then --|*** CAD: how to do in the new lookup?
			--   -- "enlive" the non-separate version of the feature
			--   local_type := current_type.local_from_separate
			--   local_run_feature := local_type.live_type.get_feature(name)
			--end
		ensure
			adapted
		end

	adapted: BOOLEAN
			--|*** unsused! (or assertion only?)

feature {}
	put_current_thread_in (buffer: STRING) is
		require
			smart_eiffel.scoop
		do
			cpp.current_thread_for(current_type.type, buffer)
		end

feature {} -- Some SCOOP helpers:
	c_define_body is
			-- Produce C code for the body definition.
		require
			cpp.on_c
		deferred
		ensure
			cpp.on_c
		end

	in_scoop_wrapper: BOOLEAN
			-- True when defining SCOOP wrap() and unwrap() functions

invariant
	current_type /= Void

	name /= Void

	base_feature /= Void

end -- class RUN_FEATURE
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
