-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
deferred class SEPARATE_TYPE_MARK
	--
	-- Handling of the "separate FOO" type mark.
	--
	-- The subclasses of SEPARATE_TYPE_MARK are the different kinds of processors:
	--  - THREAD_TYPE_MARK         processors implemented as Threads
	--  - TYPE_PROXY          processors implemented as Proxies
	--  - TYPE_SERVER         processors implemented as Servers
	--
	-- (see scoop.html)
	--
	-- The compilation either compiles all the types in (with the CCF read at
	-- runtime, acting as factory parameter), or only the relevant types if
	-- the CCF is used at compile-time.

inherit
	TYPE_MARK

feature {ANY}
	is_separate: BOOLEAN is True

	start_position: POSITION
			-- Of first letter of keyword `separate'.

	written_name: HASHED_STRING

	mapped: TYPE_MARK
			-- Corresponding mapped type (usualy the reference type).

	run_type: TYPE_MARK
			-- When runnable only.

	is_reference: BOOLEAN is True

	is_expanded: BOOLEAN is False

	live_type: LIVE_TYPE is
		do
			Result := type.live_type
		end

	is_static: BOOLEAN is
		do
			Result := mapped.is_static
		end

	local_from_separate: TYPE_MARK is
		do
			Result := mapped
		end

	jvm_method_flags: INTEGER is 17

	actual_separate (destination: TYPE_MARK): TYPE_MARK is
		do
			Result := destination
		end

	is_generic: BOOLEAN is
		do
			Result := mapped.is_generic
		end

	generic_list: ARRAY[TYPE_MARK] is
		do
			Result := mapped.generic_list
		end

	is_user_expanded: BOOLEAN is False

	is_dummy_expanded: BOOLEAN is False

	id: INTEGER is
		do
			Result := live_type.id
		end

	type: TYPE is
		do
			Result := smart_eiffel.get_type(Current)
		end
	
	to_static (t: TYPE): like Current is
		local
			tm: TYPE_MARK
		do
			if static_memory = Void then
				check
					run_type = Void
				end
				-- to allow sharing of static_memory
				create {HASHED_DICTIONARY[like Current, TYPE]} static_memory.make
			end
			Result := static_memory.reference_at(t)
			if Result = Void then
				tm := mapped.to_static(t)
				if run_type = Void then
					Result := Current
					run_type := tm
				else
					Result := twin
					Result.set_run_type(tm)
				end
				static_memory.add(Result, t)
			end
		end

	signature_resolve_in (new_type: TYPE): TYPE is
		local
			tm: TYPE_MARK; res: like Current
		do
			tm := mapped.signature_resolve_in(new_type).canonical_type_mark --|*** memory leak
			res := twin
			res.set_run_type(tm)
			Result := res.type
		end

	class_text_name: CLASS_NAME is
		do
			Result := mapped.class_text_name
		end

	declaration_type: like Current is
		do
			Result := declaration_type_memory
			if Result = Void then
				if mapped.is_static then
					check
						mapped.declaration_type = mapped
					end
					Result := Current
				else
					check
						mapped.declaration_type /= mapped
					end
					Result := twin
					Result.make(start_position, mapped.declaration_type)
				end
				declaration_type_memory := Result
			end
		end

feature {TYPE}
	valid_redefinition_of (other: TYPE_MARK; current_type, other_type: TYPE): BOOLEAN is
		do
			Result := True --|*** not_yet_implemented
--|*** if other.is_separate then
--|*** Result := mapped = other or else mapped.is_conform_to(other)
--|*** end
		end

	can_be_assigned_to (other: TYPE_MARK; current_type, other_type: TYPE): BOOLEAN is
		do
			Result := True --|*** not_yet_implemented
--|*** if other.is_separate then
--|*** Result := mapped.type.is_a(o, tp, op) --*** perf: type lookup
--|*** end
		end

feature {TYPE_MARK}
	set_start_position (sp: like start_position) is
		do
			start_position := sp
		end

feature {ANY} -- JVM: not implemented!
	jvm_descriptor_in (str: STRING) is
		do
			not_yet_implemented
		end

	jvm_target_descriptor_in (str: STRING) is
		do
			not_yet_implemented
		end

	jvm_return_code is
		do
			not_yet_implemented
		end

	jvm_push_local (offset: INTEGER) is
		do
			not_yet_implemented
		end

	jvm_check_class_invariant is
		do
			not_yet_implemented
		end

	jvm_push_default: INTEGER is
		do
			not_yet_implemented
		end

	jvm_write_local_creation (offset: INTEGER) is
		do
			not_yet_implemented
		end

	jvm_write_local (offset: INTEGER) is
		do
			not_yet_implemented
		end

	jvm_write_local_expanded (offset: INTEGER) is
		do
			not_yet_implemented
		end

	jvm_xnewarray is
		do
			not_yet_implemented
		end

	jvm_xastore is
		do
			not_yet_implemented
		end

	jvm_xaload is
		do
			not_yet_implemented
		end

	jvm_if_x_eq: INTEGER is
		do
			not_yet_implemented
		end

	jvm_if_x_ne: INTEGER is
		do
			not_yet_implemented
		end

	jvm_standard_is_equal is
		do
			not_yet_implemented
		end

feature {TYPE_MARK}
	set_run_type (rt: like run_type) is
		do
			run_type := rt
		end

feature {}
	static_memory: DICTIONARY[like Current, TYPE]

	declaration_type_memory: like Current

feature {TYPE_MARK}
	frozen short_ (shorted_type: TYPE) is
		do
			short_printer.hook_or(fz_separate, "separate ")
			mapped.short_(shorted_type)
		end

feature {SEPARATE_TYPE_MARK}
	make (sp: like start_position; m: like mapped) is
		require
			not sp.is_unknown
			m /= Void
		local
			mapped_wm, buffer: STRING
		do
			start_position := sp
			mapped := m -- Compute `written_name':
			mapped_wm := mapped.written_mark
			buffer := once "...This is a local once buffer..."
			buffer.copy(fz_separate)
			buffer.extend(' ')
			buffer.append(mapped_wm)
			written_name := string_aliaser.hashed_string(buffer)
		ensure
			start_position = sp
			mapped = m
		end

feature {}
	from_other (other: like Current; m: like mapped) is
		require
			other /= Void
			m.is_static
		do
			start_position := other.start_position
			written_name := other.written_name
			mapped := m
			if mapped.is_separate then
				run_type := mapped
			else
				run_type := Current
			end
		ensure
			run_type /= Void
		end

	copy_from_other (other: like Current; m: like mapped): like Current is
			-- Factory method, used to duplicate the current object.
		deferred
		end

end -- class SEPARATE_TYPE_MARK
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
