-- This file is part of SmartEiffel The GNU Eiffel Compiler Tools and Libraries.
-- See the Copyright notice at the end of this file.
--
class CLIENT_LIST
	--
	-- To store a list of clients class like : {FOO, BAR}
	--

inherit
	VISITABLE
		redefine is_equal
		end

insert
	GLOBALS
		redefine is_equal
		end

creation {ANY}
	make, omitted, merge

feature {ANY}
	start_position: POSITION
			-- Of the the opening bracket when list is really written.

	is_omitted: BOOLEAN is
		do
			Result := start_position.is_unknown
		end

	pretty (indent_level: INTEGER) is
		do
			if is_omitted then
				if pretty_printer.zen_mode then
					-- No print.
				else
					pretty_printer.put_string(once "{ANY}")
				end
			else
				if list = Void then
					if pretty_printer.parano_mode then
						pretty_printer.put_string(once "{NONE}")
					else
						pretty_printer.put_string(once "{}")
					end
				else
					pretty_printer.put_character('{')
					list.pretty(indent_level)
					pretty_printer.put_character('}')
				end
			end
		end

	gives_permission_to (cn: CLASS_NAME): BOOLEAN is
			-- Check whether the `cn' class is a member (or a subclass as
			-- well) of the `Current' client list. (No error report done here
			-- in `error_handler').
		require
			cn /= Void
		do
			if is_omitted then
				-- It is equivalent to {ANY}
				Result := True
			elseif cn.to_string = as_typed_internals then
				Result := True
			elseif cn.to_string = as_native_array_internals then
				Result := True
			elseif list = Void then
				-- It is equivalent to {NONE}
				Result := cn.to_string = as_none
			else
				Result := list.gives_permission_to(cn)
			end
		ensure
			(old (error_handler.is_empty)) implies error_handler.is_empty
		end

	gives_permission_to_any: BOOLEAN is
			-- Check whether the `Current' client list gives permission to all
			-- classes. (No error report done here in `error_handler').
		do
			if is_omitted then
				Result := True -- Because it is as : {ANY}.
			elseif list = Void then
				-- Because it is as : {NONE}.
			else
				Result := list.gives_permission_to_any
			end
		ensure
			(old (error_handler.is_empty)) implies error_handler.is_empty
		end

	gives_permission_to_none: BOOLEAN is
			-- Check whether the `Current' client list gives no permission at all
			-- (No error report done here in `error_handler').
		do
			if is_omitted then
				-- Because it is as : {ANY}.
			elseif list = Void then
				-- Because it is as : {NONE}.
				Result := True
			else
				Result := list.gives_permission_to_none
			end
		ensure
			(old (error_handler.is_empty)) implies error_handler.is_empty
		end

	is_equal (other: like Current): BOOLEAN is
		do
			Result := other = Current or else other.list = list or else list /= Void and then other.list /= Void and then list.is_equal(other.list)
		ensure then
			Result = (wider_than(other) and other.wider_than(Current))
		end

	accept (visitor: CLIENT_LIST_VISITOR) is
		do
			visitor.visit_client_list(Current)
		end

feature {ANONYMOUS_FEATURE_MIXER, CLIENT_LIST_VISITOR}
	append_in (b: STRING) is
		do
			if is_omitted then
				b.append(once "{ANY}")
			else
				if list = Void then
					b.append(once "{}")
				else
					b.extend('{')
					list.append_in(b)
					b.extend('}')
					b.extend(' ')
				end
			end
		end

feature {ANY}
	eiffel_view: STRING is
			-- The Eiffel view of the allowed classe(s) list. (Because of
			-- clients list merging, the `Current' clients list may be located
			-- on many Eiffel source files. This function is also useful to
			-- remind default abbreviated notation as omitted list or empty
			-- list.)
		local
			i: INTEGER
		do
			if eiffel_view_memory = Void then
				if is_omitted then
					eiffel_view_memory := once "{ANY}"
				elseif list = Void then
					eiffel_view_memory := once "{NONE}"
				else
					create eiffel_view_memory.make(64)
					eiffel_view_memory.extend('{')
					from
						i := 1
					until
						i > list.count
					loop
						eiffel_view_memory.append(list.item(i).to_string)
						i := i + 1
						if i <= list.count then
							eiffel_view_memory.extend(',')
							eiffel_view_memory.extend(' ')
						end
					end
					eiffel_view_memory.extend('}')
				end
			end
			Result := eiffel_view_memory
		ensure
			Result /= Void
		end

feature {CLASS_TEXT, CLIENT_LIST, CALL_SUPPORT}
	locate_in_error_handler is
			-- Add one or more related positions in the `error_handler'.
		do
			if list = Void then
				error_handler.add_position(start_position)
			else
				list.locate_in_error_handler
			end
		end

feature {CLIENT_LIST, ANONYMOUS_FEATURE_MIXER}
	wider_than (other: like Current): BOOLEAN is
		do
			if Current = other then
				Result := True
			elseif gives_permission_to_any then
				Result := True
			elseif other = Void or else other.gives_permission_to_any then
				--Result := not gives_permission_to_any -- always False
			elseif other.gives_permission_to_none then
				Result := True
			elseif gives_permission_to_none then
				--Result := not other.gives_permission_to_none -- always False
			else
				Result := list.wider_than(other.list)
			end
		end

feature {EXPORT_LIST, INHERIT_TEXT, ANONYMOUS_FEATURE_MIXER}
	merge_with (other: like Current): like Current is
			--|*** CAD: I say intersection, Dom says union. Currently it's a union, but I'm pretty sure it should be an
			--|intersection!
		require
			other /= Void
		local
			sp: POSITION
		do
			if Current = other then
				Result := Current
			elseif gives_permission_to_any then
				Result := Current
			elseif other.gives_permission_to_any then
				Result := other
			elseif gives_permission_to_none then
				Result := other
			elseif other.gives_permission_to_none then
				Result := Current
			else
				-- We really have to create a new one :
				sp := start_position
				if sp.is_unknown then
					sp := other.start_position
				end
				create Result.merge(sp, list, other.list)
			end
		end

feature {CLIENT_LIST, CLIENT_LIST_VISITOR}
	list: CLASS_NAME_LIST

feature {}
	make (sp: like start_position; l: like list) is
			-- When the client list is really written.
			--
			-- Note : {NONE} has the same meaning as {}.
		require
			not sp.is_unknown
		do
			start_position := sp
			list := l
			debug
				if eiffel_view /= Void then
				end
			end
		ensure
			start_position = sp
			list = l
		end

	omitted is
			-- When the client list is omitted. (Remind that when the
			-- client list is omitted, it is like {ANY}.)
		do
		end

	merge (sp: like start_position; l1, l2: like list) is
		require
			not sp.is_unknown
		do
			start_position := sp
			create list.merge(l1, l2)
			debug
				eiffel_view_memory := Void
				if eiffel_view /= Void then
				end
			end
		end

	eiffel_view_memory: STRING
			-- To cache the Result of `eiffel_view'.

end -- class CLIENT_LIST
--
-- ------------------------------------------------------------------------------------------------------------------------------
-- Copyright notice below. Please read.
--
-- SmartEiffel is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License,
-- as published by the Free Software Foundation; either version 2, or (at your option) any later version.
-- SmartEiffel is distributed in the hope that it will be useful but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have
-- received a copy of the GNU General Public License along with SmartEiffel; see the file COPYING. If not, write to the Free
-- Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
--
-- Copyright(C) 1994-2002: INRIA - LORIA (INRIA Lorraine) - ESIAL U.H.P.       - University of Nancy 1 - FRANCE
-- Copyright(C) 2003-2004: INRIA - LORIA (INRIA Lorraine) - I.U.T. Charlemagne - University of Nancy 2 - FRANCE
--
-- Authors: Dominique COLNET, Philippe RIBET, Cyril ADRIAN, Vincent CROIZIER, Frederic MERIZEN
--
-- http://SmartEiffel.loria.fr - SmartEiffel@loria.fr
-- ------------------------------------------------------------------------------------------------------------------------------
